# Copyright 2014 Eugene Frolov <eugene@frolov.net.ru>
#
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import abc
import collections
import inspect
import posixpath
import re

from restalchemy.api import actions
from restalchemy.api import constants
from restalchemy.api import controllers
from restalchemy.common import exceptions as exc
from restalchemy.openapi import constants as oa_c
from restalchemy.openapi import parse
from restalchemy.openapi import structures

# RA HTTP methods
GET = constants.GET
FILTER = constants.FILTER
CREATE = constants.CREATE
UPDATE = constants.UPDATE
DELETE = constants.DELETE

# Other HTTP methods
PUT = constants.PUT
POST = constants.POST

# route constants
COLLECTION_ROUTE = 1
RESOURCE_ROUTE = 2

SPECIAL_SYMBOLS = re.compile(r"[/{}!@#$%^&*())\[\]:,./<>?\|`\~\-\=\_\+]+")


class BaseRoute(metaclass=abc.ABCMeta):
    __controller__ = None
    __allow_methods__ = []
    __tags__ = None  # list of strings or dicts with name and description

    def __init__(self, req):
        super(BaseRoute, self).__init__()
        self._req = req

    @classmethod
    def get_controller_class(cls):
        return cls.__controller__

    @classmethod
    def get_controller(cls, *args, **kwargs):
        return cls.get_controller_class()(*args, **kwargs)

    @classmethod
    def get_allow_methods(cls):
        return cls.__allow_methods__

    @abc.abstractmethod
    def do(self, **kwargs):
        pass

    @classmethod
    def openapi_tags(cls, for_paths=False):
        if not cls.__tags__:
            controller = cls.get_controller_class()
            if controller:
                resource = controller.get_resource()
                if resource:
                    if for_paths:
                        return [resource.get_model().__name__]
                    else:
                        return [
                            {
                                "name": resource.get_model().__name__,
                                "description": "",
                            }
                        ]
            return []
        res = []
        if isinstance(cls.__tags__, list):
            for tag in cls.__tags__:
                if isinstance(tag, str):
                    if for_paths:
                        res.append(tag)
                    else:
                        res.append({"name": tag, "description": ""})
                elif isinstance(tag, structures.OpenApiTag):
                    if for_paths:
                        res.append(tag.name)
                    else:
                        res.append(tag.build())
        return res

    @staticmethod
    def restore_path_info(req):
        """Restore initial values for path_info and script_name to not raise
        UnsupportedHttpMethod exception when RetryOnErrorMiddleware triggers.
        """

        req.path_info, req.script_name = req.script_name, req.path_info


class Route(BaseRoute):
    __controller__ = None
    __allow_methods__ = [GET, CREATE, UPDATE, DELETE, FILTER]

    @classmethod
    def is_resource_route(cls):
        return False

    @classmethod
    def is_collection_route(cls):
        return True

    @classmethod
    def get_attr_safe(cls, name, the_class):
        try:
            attr = getattr(cls, name.replace("-", "_"))
            if not (inspect.isclass(attr) and issubclass(attr, the_class)):
                raise exc.IncorrectRouteAttributeClass(route=attr)
            return attr
        except AttributeError:
            raise exc.IncorrectRouteAttribute(route=cls, attr=name)

    @classmethod
    def get_route(cls, name):
        return cls.get_attr_safe(name, Route)

    @classmethod
    def get_action(cls, name):
        return cls.get_attr_safe(name, Action)

    @classmethod
    def is_route(cls, name):
        try:
            cls.get_route(name)
            return True
        except (exc.IncorrectRouteAttributeClass, exc.IncorrectRouteAttribute):
            return False

    @classmethod
    def is_action(cls, name):
        try:
            cls.get_action(name)
            return True
        except (exc.IncorrectRouteAttributeClass, exc.IncorrectRouteAttribute):
            return False

    @classmethod
    def get_routes(cls):
        return filter(lambda x: cls.is_route(x), dir(cls))

    @classmethod
    def get_action_names(cls):
        return filter(lambda x: cls.is_action(x), dir(cls))

    @classmethod
    def get_actions_by_names(cls, names):
        return [getattr(cls.get_controller_class(), name) for name in names]

    @classmethod
    def check_allow_methods(cls, *args):
        for method in args:
            if method not in cls.__allow_methods__:
                return False
        return True

    def get_method_by_route_type(self, route_type):
        if route_type == COLLECTION_ROUTE:
            mapping = {GET: FILTER, POST: CREATE}
        else:
            mapping = {GET: GET, PUT: UPDATE, DELETE: DELETE}
        try:
            return mapping[self._req.method]
        except KeyError:
            raise exc.UnsupportedHttpMethod(method=self._req.method)

    def _build_openapi_routes_specification(
        self, route_names, current_path, parameters=None
    ):
        for name in route_names:
            next_route = self.get_route(name)(self._req)
            route_path = posixpath.join(current_path, name + "/")
            yield next_route.build_openapi_specification(route_path, parameters)

    @staticmethod
    def _build_operation_id(method, current_path):
        operation_id = "{}_{}".format(method.capitalize(), current_path)
        operation_id = re.sub(SPECIAL_SYMBOLS, "_", operation_id)
        operation_id = operation_id.replace("__", "_")
        operation_id = operation_id.rstrip("_")
        return operation_id

    def _build_openapi_method_specification(
        self, method, parameters=None, current_path="/", controller=None
    ):
        controller = controller or self.get_controller(request=self._req)
        if isinstance(method, actions.ActionHandler):
            method_name = method.name
        else:
            method = getattr(controller, method.lower())
            method_name = getattr(method, "method", method.__name__.upper())
        openapi_schema = getattr(method, "openapi_schema", None)
        if openapi_schema:
            summary = openapi_schema.summary
            responses = openapi_schema.responses
            params = openapi_schema.parameters
            operation_id = openapi_schema.operation_id
            r_body = openapi_schema.request_body
            tags = openapi_schema.tags
        else:
            summary, responses, params, operation_id, r_body, tags = [None] * 6

        parsed_doc = parse.parse_docstring(method.__doc__)
        res = controller.get_resource()
        model_name = res.get_model().__name__ if res else "Unknown"
        summary_name = (
            model_name
            if res
            else "{} {}".format(self.__class__.__name__, controller.__class__.__name__)
        )
        if method_name == constants.FILTER:
            summary_name += "s"
        if not summary:
            summary = parsed_doc["short_description"]
            summary = summary or "%s %s" % (
                (
                    method_name.capitalize()
                    if method_name != constants.FILTER
                    else constants.GET.capitalize()
                ),
                summary_name,
            )

        # Fill responses
        if not responses:
            responses = parsed_doc["returns"]
        method_model_name = "{}_{}".format(model_name, method_name.capitalize())
        if not responses:
            if method_name in [constants.UPDATE, constants.GET]:
                if res:
                    responses = oa_c.build_openapi_get_update_response(
                        method_model_name
                    )
                else:
                    responses = oa_c.OPENAPI_FILTER_RESPONSE
            elif method_name == constants.FILTER:
                if res:
                    responses = oa_c.build_openapi_list_model_response(
                        method_model_name
                    )
                else:
                    responses = oa_c.OPENAPI_FILTER_RESPONSE
            elif method_name == constants.CREATE:
                responses = oa_c.build_openapi_create_response(method_model_name)
            elif method_name == constants.DELETE:
                responses = oa_c.build_openapi_delete_response(method_model_name)
            else:
                responses = oa_c.OPENAPI_DEFAULT_RESPONSE

        # fill url parameters
        if not params:
            params = parsed_doc["params"]
        if not params and parameters:
            path_params = re.findall(r"\{(.*?)\}", current_path)
            for path_param in path_params:
                param = (
                    parameters.get("components", {})
                    .get("parameters", {})
                    .get(path_param, {})
                )
                if param:
                    param["name"] = path_param
                    params.append(param)
            if res and self.is_collection_route() and (method_name == constants.FILTER):
                for name, prop in res.get_fields_by_request(self._req):
                    param = (
                        parameters.get("components", {})
                        .get("parameters", {})
                        .get(prop.api_name, {})
                    )
                    # array or object not supported in query
                    if param.get("schema", {}).get("type", "") in [
                        "array",
                        "object",
                    ]:
                        continue
                    # sum type parameter not implemented in Go client
                    # NOTE(v.burygin): maybe in openapi spec somewhere
                    # we should make options
                    # if param.get("schema", {}).get("oneOf"):
                    #     continue
                    if param:
                        params.append(param)

        if not operation_id:
            operation_id = self._build_operation_id(method_name, current_path)
        result = {
            "summary": summary,
            "tags": tags or self.openapi_tags(for_paths=True),
            "parameters": params,
            "responses": responses,
            "operationId": operation_id,
        }

        # Fill request_body
        if r_body:
            result["requestBody"] = r_body
        elif method_name in [constants.CREATE, constants.UPDATE]:
            result["requestBody"] = oa_c.build_openapi_json_req_body(method_model_name)

        return result

    def build_openapi_specification(self, current_path="/", parameters=None):

        paths_result = collections.defaultdict(dict)
        schemas_result = collections.defaultdict(dict)

        if self.__class__ == OpenApiSpecificationRoute:
            ctr = self.get_controller(request=self._req)
            versions = ctr.filter({})
            for version in versions:
                resource_path = posixpath.join(current_path, version)
                paths_result[resource_path][GET.lower()] = (
                    self._build_openapi_method_specification(
                        GET, parameters, current_path
                    )
                )

        openapi_collection_methods = {GET: FILTER, POST: CREATE}
        openapi_resource_methods = {GET: GET, PUT: UPDATE, DELETE: DELETE}

        # build specification for collection methods
        for http_method, ra_method in openapi_collection_methods.items():
            if self.check_allow_methods(ra_method):
                paths_result[current_path][http_method.lower()] = (
                    self._build_openapi_method_specification(
                        ra_method, parameters, current_path
                    )
                )
            routes = [
                r for r in self.get_routes() if self.get_route(r).is_collection_route()
            ]
            route_gen = self._build_openapi_routes_specification(
                routes, current_path, parameters
            )
            for p, s in route_gen:
                paths_result.update(p)
                schemas_result.update(s)

        # build specification for resource methods
        resource = self.get_controller(request=self._req).get_resource()
        if resource is not None:
            model = resource.get_model()
            try:
                id_prop_struct = model.get_id_property()
                id_name = list(id_prop_struct)[0]
            except TypeError:
                id_name = ""
            id_parameter_name = "{%s%s}" % (
                model.__name__,
                id_name.capitalize(),
            )
            resource_path = posixpath.join(current_path, id_parameter_name)
            for http_method, ra_method in openapi_resource_methods.items():
                if self.check_allow_methods(ra_method):
                    paths_result[resource_path][http_method.lower()] = (
                        self._build_openapi_method_specification(
                            ra_method, parameters, resource_path
                        )
                    )
            routes = [
                route_name
                for route_name in self.get_routes()
                if self.get_route(route_name).is_resource_route()
            ]
            route_gen = self._build_openapi_routes_specification(
                routes, resource_path, parameters
            )
            for p, s in route_gen:
                paths_result.update(p)
                schemas_result.update(s)

            action_names = [r for r in self.get_action_names()]
            if action_names:
                route_actions = self.get_actions_by_names(action_names)
                for route_action in route_actions:
                    is_invoke = getattr(self, route_action.name).is_invoke()
                    if is_invoke:
                        action_path = posixpath.join(
                            current_path,
                            id_parameter_name,
                            "actions",
                            route_action.name,
                            "invoke",
                        )
                    else:
                        action_path = posixpath.join(
                            current_path,
                            id_parameter_name,
                            "actions",
                            route_action.name,
                        )
                    paths_result[action_path][route_action.method.lower()] = (
                        self._build_openapi_method_specification(
                            route_action, parameters, action_path
                        )
                    )
        return paths_result, schemas_result

    @classmethod
    def build_resource_map(cls, root_route, path_stack=None):
        path_stack = path_stack or []

        def build_path(resource, path_stack):
            path_stack = path_stack[:]
            path_stack.append(resource)
            return path_stack

        def build(route, path_stack):
            result = []

            controller = route.get_controller_class()
            resource = controller.get_resource()

            if route.check_allow_methods(GET):
                route_path_stack = build_path(resource, path_stack)
                result.append((resource, controller, route_path_stack))

            for name in route.get_routes():
                new_route = route.get_route(name)
                new_path = (
                    build_path(resource, path_stack)
                    if new_route.is_resource_route()
                    else path_stack[:]
                )
                new_path.append(name)
                result += build(route.get_route(name), new_path)

            return result

        class ResourceLocator(object):
            def __init__(self, path_stack, controller):
                self.path_stack = path_stack
                self._controller = controller

            def is_your_uri(self, uri):
                uri_pieces = uri.split("/")[1:]
                if len(uri_pieces) == len(self.path_stack):
                    for piece1, piece2 in zip(uri_pieces, self.path_stack):
                        if (isinstance(piece2, str) and piece1 == piece2) or (
                            not isinstance(piece2, str)
                        ):
                            continue
                        return False
                    return True
                else:
                    return False

            def get_parent_model(self, parent_type, model, resource):
                if hasattr(resource, "get_parent_model"):
                    return resource.get_parent_model(model)
                models = []
                for name, prop in resource.get_fields():
                    if prop.get_type() is parent_type:
                        models.append(getattr(model, name))
                if len(models) == 1:
                    return models[0]
                raise ValueError(
                    "Can't find resource %s. Please "
                    "implement get_parent_model in your model "
                    "(%s)" % (parent_type, type(model))
                )

            def get_uri(self, model):
                resource = self.path_stack[-1]
                path = str(resource.get_resource_id(model))
                for piece in reversed(self.path_stack[:-1]):
                    if isinstance(piece, str):
                        path = posixpath.join(piece, path)
                    else:
                        model = self.get_parent_model(
                            piece.get_model(), model, resource
                        )
                        resource = piece
                        path = posixpath.join(resource.get_resource_id(model), path)
                # FIXME(Eugene Frolov): Header must be string. Not unicode.
                return str(posixpath.join("/", path))

            def get_resource(self, request, uri, parent_resource=None):
                uuid = posixpath.basename(uri)
                if parent_resource:
                    return self._controller(request=request).get_resource_by_uuid(
                        uuid=uuid, parent_resource=parent_resource
                    )
                return self._controller(request=request).get_resource_by_uuid(uuid)

        resource_map = {}

        for res, controller, stack in build(root_route, path_stack):
            resource_map[res] = ResourceLocator(stack, controller)

        return resource_map

    def do(self, parent_resource=None, **kwargs):
        super(Route, self).do(**kwargs)

        # TODO(Eugene Frolov): Check the possibility to pass to the method
        #                      specified in a route.
        name, path = self._req.path_info_pop(), self._req.path_info_peek()

        if path is None:
            # Collection or Resource method
            ctrl_method = (
                self.get_method_by_route_type(COLLECTION_ROUTE)
                if name == ""
                else self.get_method_by_route_type(RESOURCE_ROUTE)
            )
            if self.check_allow_methods(ctrl_method):
                worker = self.get_controller(request=self._req)
                self.restore_path_info(self._req)
                if name == "":
                    # Collection method
                    return worker.do_collection(parent_resource)

                # Resource method
                return worker.do_resource(name, parent_resource)
            else:
                raise exc.UnsupportedHttpMethod(method=ctrl_method)

        elif name != "" and path is not None and self.is_route(name):
            # Next route
            route = self.get_route(name)
            if route.is_resource_route():
                raise exc.ResourceNotFoundError(resource=name, path=path)
            worker = route(self._req)
            return worker.do(parent_resource)

        elif name != "" and path == "actions":
            # Action route
            worker = self.get_controller(self._req)
            resource = worker.get_resource_by_uuid(name, parent_resource)
            self._req.path_info_pop()
            action_name = self._req.path_info_peek()
            action = self.get_action(action_name)
            worker = action(self._req)
            return worker.do(resource=resource)

        elif name != "" and path is not None:
            # Intermediate resource route
            worker = self.get_controller(self._req)
            parent_resource = worker.get_resource_by_uuid(name, parent_resource)
            name, path = self._req.path_info_pop(), self._req.path_info_peek()
            route = self.get_route(name)
            if route.is_collection_route():
                raise exc.CollectionNotFoundError(collection=name, path=path)
            worker = route(self._req)
            return worker.do(parent_resource)

        else:
            # Other
            raise exc.NotFoundError(path=path)


def route(route_class, resource_route=False):
    @classmethod
    def is_resource_route(cls):
        return resource_route

    @classmethod
    def is_collection_route(cls):
        return not resource_route

    route_class.is_resource_route = is_resource_route
    route_class.is_collection_route = is_collection_route

    return route_class


class Action(BaseRoute):
    __controller__ = None
    __allow_methods__ = [GET]

    @classmethod
    def is_invoke(cls):
        return False

    def do(self, resource, **kwargs):
        super(Action, self).do(**kwargs)

        method = self._req.method
        action_name = self._req.path_info_pop().replace("-", "_")
        invoke_info = self._req.path_info_pop()
        if invoke_info == "invoke":
            invoke = True
        elif invoke_info is None:
            invoke = False
        else:
            raise exc.UnsupportedMethod(method=invoke_info, object_name=action_name)
        controller = self.get_controller(self._req)
        action = getattr(controller, action_name)
        content_type = self._req.headers.get("Content-Type")
        if content_type == constants.DEFAULT_CONTENT_TYPE:
            body = self._req.body
            if body:
                # NOTE(v.burygin): update kwargs by request body
                packer = controller.get_packer(content_type)
                # set packer _rt None for sending any fields to action method
                packer._rt = None
                kwargs.update(**packer.unpack(value=body))
        else:
            kwargs.update(**self._req.api_context.params)
        if (method in [GET, POST, PUT] and self.is_invoke() and invoke) or (
            method == GET and not self.is_invoke() and not invoke
        ):
            action_method = getattr(action, "do_%s" % method.lower())
            self.restore_path_info(self._req)
            return action_method(controller=controller, resource=resource, **kwargs)
        else:
            raise exc.IncorrectActionCall(action=action, method=method)


def action(action_class, invoke=False):
    @classmethod
    def is_invoke(cls):
        return invoke

    action_class.is_invoke = is_invoke
    return action_class


class OpenApiSpecificationRoute(Route):
    __controller__ = controllers.OpenApiSpecificationController
    __allow_methods__ = [FILTER, GET]


class RootRoute(Route):
    __controller__ = controllers.RootController
    __allow_methods__ = [FILTER, GET]

    specifications = route(OpenApiSpecificationRoute)



# Copyright (c) 2014 Mirantis, Inc
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
# implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import threading


class MetaSingleton(type):
    """Meta Singleton

    For example:

    >>> class ConcreteSingleton(metaclass=MetaSingleton):
    ...     pass

    """

    _instance = None
    _lock = threading.Lock()

    def __call__(cls, *args, **kwargs):
        if cls._instance is None:
            with cls._lock:
                if not cls._instance:
                    cls._instance = super(MetaSingleton, cls).__call__(*args, **kwargs)
        return cls._instance


class InheritSingleton(object, metaclass=MetaSingleton):
    """Inherit Singleton

    For example:

    >>> class ConcreteSingleton(InheritSingleton):
    ...     pass
    """

    pass



# Copyright 2021 Eugene Frolov <eugene@frolov.net.ru>
#
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from restalchemy.dm import models
from restalchemy.dm import properties
from restalchemy.dm import types
from restalchemy.tests.functional import base


class FakeModel(models.Model):
    mutable_dict = properties.property(types.Dict(), required=True, mutable=True)
    mutable_list = properties.property(types.List(), default=list, mutable=True)


class DirtyPropertiesTestCase(base.BaseFunctionalTestCase):
    def test_dirty_for_mutable_property_append_values(self):
        target = FakeModel(mutable_dict={"test": "test"})

        target.mutable_dict["xxx"] = "yyyy"

        self.assertTrue(target.is_dirty())

    def test_dirty_for_mutable_property_append_deep_values(self):
        target = FakeModel(mutable_dict={"test": {}})

        target.mutable_dict["test"]["xxx"] = "yyyy"

        self.assertTrue(target.is_dirty())

    def test_dirty_for_mutable_property_change_values(self):
        target = FakeModel(mutable_dict={"test": "test"})

        target.mutable_dict["test"] = "yyyy"

        self.assertTrue(target.is_dirty())

    def test_dirty_for_mutable_property_append_to_default(self):
        target = FakeModel(mutable_dict={})

        target.mutable_list.append("test")

        self.assertTrue(target.is_dirty())



# Copyright 2016 Eugene Frolov <eugene@frolov.net.ru>
# Copyright 2021 Eugene Frolov.
#
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from restalchemy.storage.sql import migrations


class MigrationStep(migrations.AbstractMigrationStep):
    def __init__(self):
        self._depends = [""]

    @property
    def migration_id(self):
        return "9e335f87-7d66-4768-be00-05ea4cc24a89"

    def upgrade(self, session):
        sql_expr_list = [
            """
                CREATE TABLE IF NOT EXISTS batch_insert (
                    uuid CHAR(36) NOT NULL,
                    foo_field1 INT NOT NULL,
                    foo_field2 VARCHAR(255) NULL,
                PRIMARY KEY (uuid),
                UNIQUE (foo_field1));
            """,
            """
                CREATE TABLE IF NOT EXISTS batch_delete_one_pk (
                    uuid CHAR(36) NOT NULL,
                    foo_field1 INT NOT NULL,
                    foo_field2 VARCHAR(255) NOT NULL,
                PRIMARY KEY (uuid));
            """,
            """
                CREATE TABLE IF NOT EXISTS batch_delete_two_pk (
                    uuid CHAR(36) NOT NULL,
                    foo_field1 INT NOT NULL,
                    foo_field2 VARCHAR(255) NOT NULL,
                PRIMARY KEY (uuid, foo_field1));
            """,
            """
                INSERT INTO batch_delete_one_pk (
                     uuid, foo_field1, foo_field2
                ) VALUES (
                    '00000000-0000-0000-0000-000000000000', 0, '0'
                ), (
                    '00000000-0000-0000-0000-000000000001', 1, '1'
                ), (
                    '00000000-0000-0000-0000-000000000002', 2, '2'
                ), (
                    '00000000-0000-0000-0000-000000000003', 3, '3'
                )
            """,
            """
                INSERT INTO batch_delete_two_pk (
                     uuid, foo_field1, foo_field2
                ) VALUES (
                    '00000000-0000-0000-0000-000000000000', 0, '0'
                ), (
                    '00000000-0000-0000-0000-000000000001', 1, '1'
                ), (
                    '00000000-0000-0000-0000-000000000002', 2, '2'
                ), (
                    '00000000-0000-0000-0000-000000000003', 3, '3'
                )
            """,
        ]

        [session.execute(expr) for expr in sql_expr_list]

    def downgrade(self, session):
        self._delete_table_if_exists(session, "batch_insert")
        self._delete_table_if_exists(session, "batch_delete_one_pk")
        self._delete_table_if_exists(session, "batch_delete_two_pk")


migration_step = MigrationStep()



# Copyright 2016 Eugene Frolov <eugene@frolov.net.ru>
#
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from restalchemy.dm import relationships
from restalchemy.storage.sql import orm
from restalchemy.tests.functional.restapi.ra_based.microservice import models


class VM(models.VM, orm.SQLStorableMixin):
    __tablename__ = "vms"


class VMNoProcessFilters(models.VM, orm.SQLStorableMixin):
    __tablename__ = "vms"


class VMNoSort(models.VM, orm.SQLStorableMixin):
    __tablename__ = "vms"


class VMDefSort(models.VM, orm.SQLStorableMixin):
    __tablename__ = "vms"


class Port(models.Port, orm.SQLStorableMixin):
    __tablename__ = "ports"

    vm = relationships.relationship(VM, required=True)


class IpAddress(models.IpAddress, orm.SQLStorableMixin):
    __tablename__ = "ip_addresses"

    port = relationships.relationship(Port, required=True)


class Tag(models.Tag, orm.SQLStorableMixin):
    __tablename__ = "tags"

    vm = relationships.relationship(VM, required=True)



# Copyright 2019 Eugene Frolov
#
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import orjson

import mock

from restalchemy.dm import models
from restalchemy.dm import properties
from restalchemy.dm import types
from restalchemy.storage import exceptions
from restalchemy.storage.sql.dialect import exceptions as dialect_exc
from restalchemy.storage.sql import orm
from restalchemy.tests.unit import base

FAKE_VALUE_A = "FAKE_A"
FAKE_VALUE_B = "FAKE_B"
FAKE_UUID = "89d423c5-4365-4be2-bde9-2730909a9af8"

FAKE_DICT = {"key": "value", "list": [1, 2, 3], "dict": {"a": "A"}}
FAKE_DICT_JSON = orjson.dumps(FAKE_DICT).decode()
FAKE_LIST = [1, "a", None]
FAKE_LIST_JSON = orjson.dumps(FAKE_LIST).decode()


class FakeRestoreModel(models.Model, orm.SQLStorableMixin):
    __tablename__ = "fake_table"

    a = properties.property(types.String())
    b = properties.property(types.String())

    def __init__(self, args, **kwargs):
        super(FakeRestoreModel, self).__init__(*args, **kwargs)
        raise AssertionError("Init method should not be called")


class FakeRestoreModelWithUUID(FakeRestoreModel, models.ModelWithUUID):
    pass


class FakeDirtyRestoreModelWithUUID(FakeRestoreModel, models.ModelWithUUID):
    def is_dirty(self):
        return True


class TestRestoreModelTestCase(base.BaseTestCase):
    def test_init_should_not_be_called(self):
        model = FakeRestoreModel.restore_from_storage(a=FAKE_VALUE_A, b=FAKE_VALUE_B)

        self.assertEqual(model.a, FAKE_VALUE_A)
        self.assertEqual(model.b, FAKE_VALUE_B)

    def test_tablename_should_be_defined(self):
        model = type(
            "TestIncompleteRestoreModel",
            (models.Model, orm.SQLStorableMixin),
            {},
        )()

        with self.assertRaises(orm.UndefinedAttribute):
            model.get_table()


class FakeRestoreWithJSONModel(models.Model, orm.SQLStorableWithJSONFieldsMixin):
    __tablename__ = "fake_table"
    __jsonfields__ = ["a", "b"]

    a = properties.property(types.Dict())
    b = properties.property(types.List())


class TestRestoreWithJSONModelTestCase(base.BaseTestCase):
    def test_json_parsed(self):
        model = FakeRestoreWithJSONModel.restore_from_storage(
            a=FAKE_DICT_JSON, b=FAKE_LIST_JSON
        )

        self.assertEqual(model.a, FAKE_DICT)
        self.assertEqual(model.b, FAKE_LIST)

    def test_json_dumped(self):
        model = FakeRestoreWithJSONModel(a=FAKE_DICT, b=FAKE_LIST)
        prepared_data = model._get_prepared_data()

        self.assertEqual(prepared_data["a"], FAKE_DICT_JSON)
        self.assertEqual(prepared_data["b"], FAKE_LIST_JSON)

    def test_tablename_should_be_defined(self):
        model = type(
            "TestIncompleteRestoreWithJSONModel",
            (models.Model, orm.SQLStorableWithJSONFieldsMixin),
            {},
        )()

        with self.assertRaises(orm.UndefinedAttribute):
            model.restore_from_storage()
        with self.assertRaises(orm.UndefinedAttribute):
            model._get_prepared_data()


class TestSimplifyModelTestCase(base.BaseTestCase):
    def test_from_model(self):
        model = FakeRestoreModelWithUUID.restore_from_storage(
            a=FAKE_DICT_JSON, b=FAKE_LIST_JSON, uuid=FAKE_UUID
        )

        self.assertEqual(
            FakeRestoreModelWithUUID.to_simple_type(model), str(model.uuid)
        )

    def test_from_id_type(self):
        self.assertEqual(
            FakeRestoreModelWithUUID.to_simple_type(FAKE_UUID), str(FAKE_UUID)
        )


@mock.patch("restalchemy.storage.sql.engines.engine_factory")
class TestModelErrorHandlingCase(base.BaseTestCase):
    @mock.patch("restalchemy.storage.sql.tables.SQLTable.insert")
    def test_insert_model_when_unknown_error_raises(
        self, model_insert_mock, engine_factory_mock
    ):
        model_insert_mock.side_effect = dialect_exc.BaseException(
            code=0, message="Unknown error"
        )
        model = FakeDirtyRestoreModelWithUUID.restore_from_storage(
            a=FAKE_VALUE_A, b=FAKE_VALUE_B
        )
        self.assertRaises(exceptions.UnknownStorageException, model.insert)

    @mock.patch("restalchemy.storage.sql.tables.SQLTable.update")
    def test_update_model_when_unknown_error_raises(
        self, model_update_mock, engine_factory_mock
    ):
        model_update_mock.side_effect = dialect_exc.BaseException(
            code=0, message="Unknown error"
        )
        model = FakeDirtyRestoreModelWithUUID.restore_from_storage(
            a=FAKE_VALUE_A, b=FAKE_VALUE_B
        )
        self.assertRaises(exceptions.UnknownStorageException, model.update)

    @mock.patch("restalchemy.storage.sql.tables.SQLTable.delete")
    def test_delete_model_when_unknown_error_raises(
        self, model_delete_mock, engine_factory_mock
    ):
        model_delete_mock.side_effect = dialect_exc.BaseException(
            code=1213, message="Unknown error"
        )
        model = FakeDirtyRestoreModelWithUUID.restore_from_storage(
            a=FAKE_VALUE_A, b=FAKE_VALUE_B
        )
        self.assertRaises(exceptions.UnknownStorageException, model.delete)

    @mock.patch("restalchemy.storage.sql.tables.SQLTable.insert")
    def test_insert_model_when_conflict_error_raises(
        self, model_insert_mock, engine_factory_mock
    ):
        model_insert_mock.side_effect = dialect_exc.Conflict(
            code=1062, message="Conflict is found"
        )
        model = FakeDirtyRestoreModelWithUUID.restore_from_storage(
            a=FAKE_VALUE_A, b=FAKE_VALUE_B
        )
        self.assertRaises(exceptions.ConflictRecords, model.insert)

    @mock.patch("restalchemy.storage.sql.tables.SQLTable.update")
    def test_update_model_when_conflict_error_raises(
        self, model_update_mock, engine_factory_mock
    ):
        model_update_mock.side_effect = dialect_exc.Conflict(
            code=1062, message="Conflict is found"
        )
        model = FakeDirtyRestoreModelWithUUID.restore_from_storage(
            a=FAKE_VALUE_A, b=FAKE_VALUE_B
        )
        self.assertRaises(exceptions.ConflictRecords, model.update)

    @mock.patch("restalchemy.storage.sql.tables.SQLTable.insert")
    def test_insert_model_when_deadlock_error_raises(
        self, model_insert_mock, engine_factory_mock
    ):
        model_insert_mock.side_effect = dialect_exc.DeadLock(
            code=1213, message="Deadlock is found"
        )
        model = FakeDirtyRestoreModelWithUUID.restore_from_storage(
            a=FAKE_VALUE_A, b=FAKE_VALUE_B
        )
        self.assertRaises(exceptions.DeadLock, model.insert)

    @mock.patch("restalchemy.storage.sql.tables.SQLTable.update")
    def test_update_model_when_deadlock_error_raises(
        self, model_update_mock, engine_factory_mock
    ):
        model_update_mock.side_effect = dialect_exc.DeadLock(
            code=1213, message="Deadlock is found"
        )
        model = FakeDirtyRestoreModelWithUUID.restore_from_storage(
            a=FAKE_VALUE_A, b=FAKE_VALUE_B
        )
        self.assertRaises(exceptions.DeadLock, model.update)

    @mock.patch("restalchemy.storage.sql.tables.SQLTable.delete")
    def test_delete_model_when_deadlock_error_raises(
        self, model_delete_mock, engine_factory_mock
    ):
        model_delete_mock.side_effect = dialect_exc.DeadLock(
            code=1213, message="Deadlock is found"
        )
        model = FakeDirtyRestoreModelWithUUID.restore_from_storage(
            a=FAKE_VALUE_A, b=FAKE_VALUE_B
        )
        self.assertRaises(exceptions.DeadLock, model.delete)



# Copyright 2021 George Melikov.
#
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import unittest

from parameterized import parameterized

from restalchemy.storage.sql.dialect.query_builder import common
from restalchemy.storage.sql.dialect.query_builder import q
from restalchemy.tests import fixtures
from restalchemy.tests.utils import make_test_name


class TestOrderByValue(unittest.TestCase):
    def setUp(self):
        self.column = common.Column(
            "1",
            None,
            fixtures.SessionFixture(),
        )

    def test_empty_type(self):
        order = q.OrderByValue(
            self.column,
            fixtures.SessionFixture(),
        )

        self.assertEqual("`1` ASC", order.compile())

    @parameterized.expand(
        [
            ("ASC",),
            ("DESC",),
            ("ASC NULLS FIRST",),
            ("ASC NULLS LAST",),
            ("DESC NULLS FIRST",),
            ("DESC NULLS LAST",),
        ],
        name_func=make_test_name,
    )
    def test_valid_type(self, sort_dir):
        order = q.OrderByValue(
            self.column,
            sort_type=sort_dir,
            session=fixtures.SessionFixture(),
        )
        self.assertTrue(order.compile())

    def test_invalid_type(self):
        with self.assertRaises(ValueError):
            q.OrderByValue(
                self.column,
                sort_type="WRONG",
                session=fixtures.SessionFixture(),
            )

    def test_valid_type_lowercase(self):
        order = q.OrderByValue(
            self.column,
            sort_type="desc",
            session=fixtures.SessionFixture(),
        )

        self.assertEqual("`1` DESC", order.compile())



# Copyright 2014 Eugene Frolov <eugene@frolov.net.ru>
#
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import datetime
import decimal
import re
import sys
import uuid

import mock

from restalchemy.dm import types
from restalchemy.tests.unit import base

TEST_STR_VALUE = "test_value :)"
TEST_INT_AS_STR_VALUE = "1234"
TEST_INT_VALUE = 5
TEST_TYPE = "FAKE TYPE"
INCORECT_UUID = "4a775g98-eg85-4a0e-a0g0-639f0a16f4c3"
INCORECT_INT_AS_STR = "123abc"


@mock.patch("re.compile", return_value=mock.MagicMock(), autospec=True)
class BaseRegExpTypeTestCase(base.BaseTestCase):
    def _prepare_mock(self, re_mock, return_value):
        self.re_match_mock = mock.MagicMock(
            **{"match": mock.MagicMock(return_value=return_value)}
        )
        re_mock.return_value = self.re_match_mock

    def test_correct_value_if_value_is_not_none(self, re_mock):
        self._prepare_mock(re_mock, re.match("a", "a"))

        test_instance = types.BaseRegExpType("")

        self.assertTrue(test_instance.validate(TEST_STR_VALUE))
        self.re_match_mock.match.assert_called_once_with(TEST_STR_VALUE)

    def test_correct_value_if_value_is_none(self, re_mock):
        self._prepare_mock(re_mock, None)

        test_instance = types.BaseRegExpType("")

        self.assertFalse(test_instance.validate(None))

    def test_incorrect_value(self, re_mock):
        self._prepare_mock(re_mock, None)

        test_instance = types.BaseRegExpType("")

        self.assertFalse(test_instance.validate(TEST_STR_VALUE))
        self.re_match_mock.match.assert_called_once_with(TEST_STR_VALUE)


class BaseCompiledRegExpTypeTestCase(base.BaseTestCase):
    def test_correct_value_if_value_is_not_none(self):
        test_instance = types.BaseCompiledRegExpType(re.compile(r"t"))

        self.assertTrue(test_instance.validate(TEST_STR_VALUE))

    def test_correct_value_if_value_is_none(self):
        test_instance = types.BaseCompiledRegExpType(re.compile(r""))

        self.assertFalse(test_instance.validate(None))

    def test_incorrect_value(self):
        test_instance = types.BaseCompiledRegExpType(re.compile(r"-"))

        self.assertFalse(test_instance.validate(TEST_STR_VALUE))

    def test_get_pattern_from_attribute(self):
        class TestType(types.BaseCompiledRegExpTypeFromAttr):
            pattern = re.compile(r"t")

        test_instance = TestType()

        self.assertTrue(test_instance.validate(TEST_STR_VALUE))


class BaseTestCase(base.BaseTestCase):
    def __init__(self, *args, **kwargs):
        super(BaseTestCase, self).__init__(*args, **kwargs)
        self.test_instance = mock.MagicMock()
        self.test_instance.validate.configure_mock(**{"return_value": False})

    def test_correct_none_value(self):
        self.assertFalse(self.test_instance.validate(None))


class UUIDTestCase(base.BaseTestCase):
    def setUp(self):
        super(UUIDTestCase, self).setUp()
        self.test_instance = types.UUID()

    def test_uuid_correct_value(self):
        self.assertTrue(self.test_instance.validate(uuid.uuid4()))

    def test_uuid_incorrect_value(self):
        INCORECT_UUID = "4a775g98-eg85-4a0e-a0g0-639f0a16f4c3"

        self.assertFalse(self.test_instance.validate(INCORECT_UUID))

    def test_to_simple_type(self):
        TEST_UUID = uuid.uuid4()

        self.assertEqual(self.test_instance.to_simple_type(TEST_UUID), str(TEST_UUID))

    def test_from_simple_type(self):
        TEST_UUID = uuid.uuid4()

        self.assertEqual(self.test_instance.from_simple_type(str(TEST_UUID)), TEST_UUID)

    def test_from_simple_type_uuid(self):
        TEST_UUID = uuid.uuid4()

        self.assertEqual(self.test_instance.from_simple_type(TEST_UUID), TEST_UUID)


class EmailTestCase(base.BaseTestCase):
    def setUp(self):
        super(EmailTestCase, self).setUp()
        self.test_instance1 = types.Email()
        self.test_instance2 = types.Email(5, 10)
        self.test_instance3 = types.Email(5, 100, check_deliverability=True)

    def test_correct_email_value(self):
        CORRECT_EMAIL = "eugene@frolov.net.ru"

        self.assertTrue(self.test_instance1.validate(CORRECT_EMAIL))

    def test_correct_short_email_value(self):
        CORRECT_EMAIL = "a@b.c"

        self.assertTrue(self.test_instance1.validate(CORRECT_EMAIL))

    def test_incorrect_email_value(self):
        INCORRECT_EMAIL = "eugene@frolov"

        self.assertFalse(self.test_instance1.validate(INCORRECT_EMAIL))

    def test_incorrect_very_short_email_value(self):
        INCORECT_EMAIL = "x@x"

        self.assertFalse(self.test_instance1.validate(INCORECT_EMAIL))

    def test_deliverability_email_value(self):
        CORRECT_DELIVERABLE_EMAIL = "eugene@gmail.com"

        self.assertTrue(self.test_instance3.validate(CORRECT_DELIVERABLE_EMAIL))

    def test_email_to_very_long_value(self):
        INCORRECT_EMAIL = "eugene@gmail.com"

        self.assertFalse(self.test_instance2.validate(INCORRECT_EMAIL))

    def test_non_deliverability_email_value(self):
        CORRECT_NOT_DELIVERABLE_EMAIL = "eugene@frolov.incorrectzone"

        self.assertFalse(self.test_instance3.validate(CORRECT_NOT_DELIVERABLE_EMAIL))


class StringTestCase(base.BaseTestCase):
    FAKE_STRING1 = "fake!!!"
    FAKE_STRING2 = "fake!!!"

    def setUp(self):
        super(StringTestCase, self).setUp()
        self.test_instance1 = types.String(min_length=5, max_length=8)
        self.test_instance2 = types.String()

    def test_correct_value(self):
        self.assertTrue(self.test_instance1.validate(self.FAKE_STRING1))

    def test_correct_unicode_value(self):
        self.assertTrue(self.test_instance1.validate(self.FAKE_STRING2))

    def test_correct_min_value(self):
        self.assertTrue(self.test_instance1.validate(self.FAKE_STRING1[:5]))

    def test_correct_min_unicode_value(self):
        self.assertTrue(self.test_instance1.validate(self.FAKE_STRING2[:5]))

    def test_correct_max_value(self):
        self.assertTrue(self.test_instance1.validate((self.FAKE_STRING1 * 2)[:8]))

    def test_correct_max_unicode_value(self):
        self.assertTrue(self.test_instance1.validate((self.FAKE_STRING2 * 2)[:8]))

    def test_incorrect_min_value(self):
        self.assertFalse(self.test_instance1.validate(self.FAKE_STRING1[:4]))

    def test_incorrect_min_unicode_value(self):
        self.assertFalse(self.test_instance1.validate(self.FAKE_STRING1[:4]))

    def test_incorrect_max_value(self):
        self.assertFalse(self.test_instance1.validate((self.FAKE_STRING1 * 2)[:9]))

    def test_incorrect_max_unicode_value(self):
        self.assertFalse(self.test_instance1.validate((self.FAKE_STRING1 * 2)[:9]))

    def test_correct_infinity_value(self):
        self.assertTrue(self.test_instance2.validate(self.FAKE_STRING1 * 100500))

    def test_incorrect_type_validate(self):
        self.assertFalse(self.test_instance1.validate(5))


class IntegerTestCase(base.BaseTestCase):
    def setUp(self):
        super(IntegerTestCase, self).setUp()

        self.test_instance = types.Integer(0, 55)

    def test_validate_correct_value(self):
        self.assertTrue(self.test_instance.validate(30))

    def test_validate_correct_max_value(self):
        self.assertTrue(self.test_instance.validate(55))

    def test_validate_correct_min_value(self):
        self.assertTrue(self.test_instance.validate(0))

    def test_validate_incorrect_value(self):
        self.assertFalse(self.test_instance.validate("TEST_STR_VALUE"))

    def test_validate_incorrect_max_value(self):
        self.assertFalse(self.test_instance.validate(56))

    def test_validate_incorrect_min_value(self):
        self.assertFalse(self.test_instance.validate(-1))

    def test_validate_sys_max_value(self):
        test_instance = types.Integer()

        self.assertTrue(test_instance.validate(sys.maxsize))

    def test_validate_sys_min_value(self):
        test_instance = types.Integer()

        self.assertTrue(test_instance.validate(-sys.maxsize))


class FloatTestCase(base.BaseTestCase):
    def setUp(self):
        super(FloatTestCase, self).setUp()

        self.test_instance = types.Float(0.0, 55.0)

    def test_validate_correct_value(self):
        self.assertTrue(self.test_instance.validate(30.0))

    def test_validate_correct_max_value(self):
        self.assertTrue(self.test_instance.validate(55.0))

    def test_validate_correct_min_value(self):
        self.assertTrue(self.test_instance.validate(0.0))

    def test_validate_incorrect_value(self):
        self.assertFalse(self.test_instance.validate("TEST_STR_VALUE"))

    def test_validate_incorrect_max_value(self):
        self.assertFalse(self.test_instance.validate(56.0))

    def test_validate_incorrect_min_value(self):
        self.assertFalse(self.test_instance.validate(-1.0))

    def test_validate_sys_max_value(self):
        test_instance = types.Float()

        self.assertTrue(test_instance.validate(float(sys.maxsize)))

    def test_validate_sys_min_value(self):
        test_instance = types.Float()

        self.assertTrue(test_instance.validate(float(-sys.maxsize)))


class DecimalTestCase(base.BaseTestCase):
    def setUp(self):
        super().setUp()

        self.test_instance = types.Decimal(0.0, 55.0)

    def test_validate_decimal_roundup(self):
        assert types.Decimal().from_simple_type(
            "0.1"
        ) + types.Decimal().from_simple_type("0.1") + types.Decimal().from_simple_type(
            "0.1"
        ) == decimal.Decimal("0.3")

    def test_validate_correct_value(self):
        self.assertTrue(self.test_instance.validate(decimal.Decimal(30.0)))

    def test_validate_correct_max_value(self):
        self.assertTrue(self.test_instance.validate(decimal.Decimal(55.0)))

    def test_validate_correct_min_value(self):
        self.assertTrue(self.test_instance.validate(decimal.Decimal(0.0)))

    def test_validate_incorrect_value(self):
        self.assertFalse(self.test_instance.validate("TEST_STR_VALUE"))

    def test_validate_incorrect_max_value(self):
        self.assertFalse(self.test_instance.validate(decimal.Decimal(56.0)))

    def test_validate_incorrect_min_value(self):
        self.assertFalse(self.test_instance.validate(decimal.Decimal(-1.0)))

    def test_validate_sys_max_value(self):
        test_instance = types.Decimal()

        self.assertTrue(test_instance.validate(decimal.Decimal(sys.maxsize)))

    def test_validate_sys_min_value(self):
        test_instance = types.Decimal()

        self.assertTrue(test_instance.validate(decimal.Decimal(-sys.maxsize)))

    def test_validate_correct_max_decimal_places(self):
        test_instance = types.Decimal(max_decimal_places=2)

        self.assertTrue(test_instance.validate(decimal.Decimal("2.22")))

    def test_validate_incorrect_max_decimal_places(self):
        test_instance = types.Decimal(max_decimal_places=2)

        self.assertFalse(test_instance.validate(decimal.Decimal("2.222")))


class UriTestCase(BaseTestCase):
    def setUp(self):
        super(UriTestCase, self).setUp()
        self.test_instance = types.Uri()

    def test_correct_value(self):
        self.assertTrue(self.test_instance.validate("/fake/fake/" + str(uuid.uuid4())))

    def test_incorect_uuid_value(self):
        self.assertFalse(self.test_instance.validate("/fake/fake/" + INCORECT_UUID))

    def test_incorect_start_char_value(self):
        self.assertFalse(self.test_instance.validate("fake/fake/" + str(uuid.uuid4())))

    def test_incorect_start_end_value(self):
        self.assertFalse(self.test_instance.validate("/fake/fake" + str(uuid.uuid4())))


class MacTestCase(BaseTestCase):
    def setUp(self):
        super(MacTestCase, self).setUp()
        self.test_instance = types.Mac()

    def get_values(self, value):
        return [value, value.upper()]

    def test_correct_value(self):
        for value in self.get_values("05:06:07:08:ab:ff"):
            self.assertTrue(self.test_instance.validate(value))

    def test_incorrect_cahar_value(self):
        for value in self.get_values("05:06:0k:08:ab:ff"):
            self.assertFalse(self.test_instance.validate(value))

    def test_incorrect_length_value(self):
        for value in self.get_values("05:06:08:ab:ff"):
            self.assertFalse(self.test_instance.validate(value))


class BasePythonTypeTestCase(base.BaseTestCase):
    def setUp(self):
        super(BasePythonTypeTestCase, self).setUp()

        self.test_instance = types.BasePythonType(int)

    def test_validate_correct_value(self):
        self.assertTrue(self.test_instance.validate(TEST_INT_VALUE))

    def test_validate_incorrect_value(self):
        self.assertFalse(self.test_instance.validate(TEST_STR_VALUE))


class ListTestCase(base.BaseTestCase):
    def setUp(self):
        super(ListTestCase, self).setUp()

        self.test_instance = types.List()

    def test_validate_correct_value(self):
        self.assertTrue(self.test_instance.validate(list()))

    def test_validate_incorrect_value(self):
        self.assertFalse(self.test_instance.validate(TEST_STR_VALUE))


class TypedListTestCase(base.BaseTestCase):
    def setUp(self):
        super(TypedListTestCase, self).setUp()

        self.test_instance_int = types.TypedList(nested_type=types.Integer())
        self.test_instance_str = types.TypedList(nested_type=types.String())

    def test_validate_correct_value(self):
        self.assertTrue(self.test_instance_int.validate([]))
        self.assertTrue(self.test_instance_int.validate([1, 2, 3]))

    def test_validate_incorrect_value(self):
        self.assertFalse(self.test_instance_int.validate([1, 2, "3", 4]))

    def test_incorrect_nested_type(self):
        self.assertRaises(TypeError, types.TypedList, int)

    def test_from_unicode(self):
        payload = ((TEST_INT_AS_STR_VALUE, [int(TEST_INT_AS_STR_VALUE)]),)

        for value, expectedResult in payload:
            result = self.test_instance_int.from_unicode(value)
            self.assertEqual(expectedResult, result)

        payload = ((TEST_STR_VALUE, [TEST_STR_VALUE]),)

        for value, expectedResult in payload:
            result = self.test_instance_str.from_unicode(value)
            self.assertEqual(expectedResult, result)

        with self.assertRaises(TypeError):
            self.test_instance_int.from_unicode(None)
            self.test_instance_int.from_unicode(INCORECT_INT_AS_STR)


class DictTestCase(base.BaseTestCase):
    def setUp(self):
        super(DictTestCase, self).setUp()

        self.test_instance = types.Dict()

    def test_validate_correct_value(self):
        self.assertTrue(self.test_instance.validate(dict()))

    def test_validate_incorrect_value(self):
        self.assertFalse(self.test_instance.validate(TEST_STR_VALUE))


class TypedDictTestCase(base.BaseTestCase):
    def setUp(self):
        super(TypedDictTestCase, self).setUp()

        self.scheme_simple_types = {
            "int": types.Integer(),
            "str": types.String(),
        }
        self.scheme_lists = {
            "list": types.List(),
            "typed_list": types.TypedList(types.Integer()),
        }
        self.scheme_dicts = {
            "dict": types.Dict(),
            "typed_dict": types.SchemeDict(
                {
                    "sub_str": types.String(),
                    "sub_int": types.Integer(),
                }
            ),
        }
        self.scheme_dict_sublist = {
            "typed_dict_with_typed_list": types.SchemeDict(
                {"sub_list_typed": types.TypedList(types.String())}
            ),
        }
        self.scheme_dict_subdict = {
            "typed_dict_with_typed_dict": types.SchemeDict(
                {
                    "sub_dict_typed": types.SchemeDict(
                        {
                            "sub_str": types.String(),
                            "sub_int": types.Integer(),
                        }
                    )
                }
            ),
        }

    def test_schema_keys_not_string(self):
        self.assertRaises(ValueError, types.SchemeDict, {1: types.Integer()})

    def test_schema_values_not_types(self):
        self.assertRaises(ValueError, types.SchemeDict, {"1": int})

    def test_validate_simple_schema(self):
        dict_type = types.SchemeDict(scheme=self.scheme_simple_types)

        self.assertTrue(dict_type.validate({"int": 1, "str": "string"}))

    def test_validate_simple_schema_missing_item(self):
        dict_type = types.SchemeDict(scheme=self.scheme_simple_types)

        self.assertFalse(dict_type.validate({"int": 1}))
        self.assertFalse(dict_type.validate({"str": "string"}))

    def test_validate_simple_schema_extra_item(self):
        dict_type = types.SchemeDict(scheme={"int": types.Integer()})

        self.assertFalse(dict_type.validate({"int": 1, "str": "string"}))

    def test_validate_simple_schema_invalid_value(self):
        dict_type_1 = types.SchemeDict(scheme={"int": types.Integer()})
        dict_type_2 = types.SchemeDict(scheme={"str": types.String()})

        self.assertFalse(dict_type_1.validate({"int": "1"}))
        self.assertFalse(dict_type_2.validate({"str": None}))

    def test_validate_schema_with_lists(self):
        schema = {"mixed_list": types.List()}
        schema.update(self.scheme_simple_types)
        schema.update(self.scheme_lists)

        dict_type = types.SchemeDict(scheme=schema)

        self.assertTrue(
            dict_type.validate(
                {
                    "int": 1,
                    "str": "string",
                    "list": [1, 2, 3],
                    "mixed_list": [1, "a", None],
                    "typed_list": [1, 2, 3],
                }
            )
        )

    def test_validate_schema_incorrect_typed_list_value(self):
        schema = {"typed_list": types.TypedList(types.Integer())}

        dict_type = types.SchemeDict(scheme=schema)

        self.assertFalse(dict_type.validate({"typed_list": [1, "2", 3]}))
        self.assertFalse(dict_type.validate({"typed_list": [None, 2, 3]}))
        self.assertFalse(dict_type.validate({"typed_list": [1, 2, {}]}))

    def test_validate_schema_with_dicts(self):
        schema = {}
        schema.update(self.scheme_simple_types)
        schema.update(self.scheme_dicts)

        dict_type = types.SchemeDict(scheme=schema)

        self.assertTrue(
            dict_type.validate(
                {
                    "int": 1,
                    "str": "string",
                    "dict": {"1": 1, "2": "a", "z": 3},
                    "typed_dict": {
                        "sub_str": "string",
                        "sub_int": 42,
                    },
                }
            )
        )

    def test_validate_schema_subdict_missing_item(self):
        schema = {}
        schema.update(self.scheme_simple_types)
        schema.update(self.scheme_dicts)

        dict_type = types.SchemeDict(scheme=schema)

        self.assertFalse(
            dict_type.validate(
                {
                    "int": 1,
                    "str": "string",
                    "dict": {1: 1, 2: "a", "z": 3},
                    "typed_dict": {
                        "sub_str": "string",
                    },
                }
            )
        )

    def test_validate_complex_schema(self):
        schema = {}
        schema.update(self.scheme_simple_types)
        schema.update(self.scheme_lists)
        schema.update(self.scheme_dicts)
        schema.update(self.scheme_dict_sublist)
        schema.update(self.scheme_dict_subdict)

        dict_type = types.SchemeDict(scheme=schema)

        self.assertTrue(
            dict_type.validate(
                {
                    "int": 1,
                    "str": "string",
                    "list": ["a"],
                    "typed_list": [],
                    "dict": {},
                    "typed_dict": {"sub_str": "string", "sub_int": 42},
                    "typed_dict_with_typed_list": {"sub_list_typed": ["s"]},
                    "typed_dict_with_typed_dict": {
                        "sub_dict_typed": {
                            "sub_str": "string_2",
                            "sub_int": -5,
                        }
                    },
                }
            )
        )

    def test_scheme_dict_to_simple_type(self):
        val = 7
        ret_val = 9
        test_obj = self

        class TestType(types.Integer):
            def to_simple_type(self, data_val):
                test_obj.assertIs(val, data_val)
                return ret_val

        scheme = {
            "key": TestType(),
        }
        data = {
            "key": val,
        }
        expect_data = {"key": ret_val}

        res = types.SchemeDict(scheme=scheme).to_simple_type(data)
        self.assertEqual(expect_data, res)


class UTCDateTimeTestCase(base.BaseTestCase):
    def setUp(self):
        super(UTCDateTimeTestCase, self).setUp()

        self.test_instance = types.UTCDateTime()

    def test_validate_correct_value_with_explicit_utc_tz(self):
        self.assertTrue(
            self.test_instance.validate(datetime.datetime.now(datetime.timezone.utc))
        )

    def test_validate_correct_value(self):
        self.assertTrue(self.test_instance.validate(datetime.datetime.utcnow()))

    def test_validate_incorrect_value_type(self):
        self.assertFalse(self.test_instance.validate(TEST_STR_VALUE))

    def test_validate_incorrect_value_tzinfo(self):
        self.assertFalse(
            self.test_instance.validate(
                datetime.datetime.utcnow().replace(tzinfo=datetime.tzinfo())
            )
        )

    def test_zero_microseconds(self):
        dt = datetime.datetime(2020, 3, 13, 11, 3, 25)
        expected = "2020-03-13 11:03:25.000000"
        dt_type = types.UTCDateTime()

        result = dt_type.to_simple_type(dt)

        self.assertEqual(result, expected)

    def test_openapi_format(self):
        dt = datetime.datetime(2020, 3, 13, 11, 3, 25, 123)

        expected = "2020-03-13T11:03:25.000123Z"
        dt_type = types.UTCDateTime()

        result = dt_type.dump_value(dt)

        self.assertEqual(result, expected)

        result = dt_type.from_simple_type(expected)
        self.assertEqual(result, dt)


class UTCDateTimeZTestCase(base.BaseTestCase):
    def setUp(self):
        super(UTCDateTimeZTestCase, self).setUp()

        self.test_instance = types.UTCDateTimeZ()

    def test_validate_correct_value_with_explicit_utc_tz(self):
        self.assertTrue(
            self.test_instance.validate(datetime.datetime.now(datetime.timezone.utc))
        )

    def test_validate_incorrect_value_wo_tz(self):
        self.assertFalse(self.test_instance.validate(datetime.datetime.now()))

    def test_validate_from_simple_type_wo_tz(self):
        dt = datetime.datetime(2020, 3, 13, 11, 3, 25)

        result = types.UTCDateTimeZ().from_simple_type(dt)

        self.assertEqual(result, dt.replace(tzinfo=datetime.timezone.utc))
        self.assertIsNone(dt.tzinfo)
        self.assertEqual(result.tzinfo, datetime.timezone.utc)

    def test_validate_from_simple_type_with_tz(self):
        dtz = datetime.datetime.fromisoformat("2020-03-13T11:03:25+03:00")
        expected_utc = "2020-03-13 08:03:25.000000"

        result = types.UTCDateTimeZ().from_simple_type(dtz)

        self.assertEqual(
            dtz.tzinfo, datetime.timezone(datetime.timedelta(seconds=10800))
        )
        self.assertEqual(result.tzinfo, datetime.timezone.utc)
        self.assertEqual(result, dtz.astimezone(datetime.timezone.utc))
        self.assertEqual(types.UTCDateTimeZ().to_simple_type(result), expected_utc)


class EnumTestCase(base.BaseTestCase):
    def setUp(self):
        super(EnumTestCase, self).setUp()

        self.test_instance = types.Enum([1, 2, 3])

    def test_validate_correct_value(self):
        self.assertTrue(self.test_instance.validate(1))

    def test_validate_incorrect_value(self):
        self.assertFalse(self.test_instance.validate(4))


class AllowNoneTestCase(base.BaseTestCase):
    def setUp(self):
        super(AllowNoneTestCase, self).setUp()

        self.test_instance = types.AllowNone(types.String())

    def test_from_unicode(self):
        s = "blah273 2 s3d3"
        self.assertEqual(self.test_instance.from_unicode(s), s)
        self.assertIsNone(self.test_instance.from_unicode("null"))

    def test_validate_correct_value(self):
        self.assertTrue(self.test_instance.validate(None))
        self.assertTrue(self.test_instance.validate("string"))

    def test_validate_incorrect_value(self):
        self.assertFalse(self.test_instance.validate(4))


class AnySimpleTypeTestCase(base.BaseTestCase):
    def setUp(self):
        super().setUp()
        self.test_instance = types.AnySimpleType()

    def test_validate_correct_value(self):
        payload = (
            1,
            1.0,
            "foo",
            [1, 2, 3],
            {"foo": "bar"},
            True,
        )

        for value in payload:
            self.assertTrue(self.test_instance.validate(value))

    def test_validate_incorrect_value(self):
        payload = (
            None,
            object(),
            datetime.datetime.utcnow(),
        )

        for value in payload:
            self.assertFalse(self.test_instance.validate(value))

    def test_from_unicode_str(self):
        payload = (
            ("123", 123),
            ("1.5", 1.5),
            ("true", True),
            ('"foo"', "foo"),
            ("[1, 2, 3]", [1, 2, 3]),
            ('{"foo": "bar"}', {"foo": "bar"}),
        )

        for value, expected in payload:
            self.assertEqual(self.test_instance.from_unicode(value), expected)

    def test_from_unicode_bytes(self):
        payload = (
            (b"123", 123),
            (b"1.5", 1.5),
            (b"true", True),
            (b'"foo"', "foo"),
            (b"[1, 2, 3]", [1, 2, 3]),
            (b'{"foo": "bar"}', {"foo": "bar"}),
        )

        for value, expected in payload:
            self.assertEqual(self.test_instance.from_unicode(value), expected)

    def test_from_unicode_invalid_json(self):
        payload = (
            "foo",
            "{",
            "[1,",
        )

        for value in payload:
            with self.assertRaises(TypeError):
                self.test_instance.from_unicode(value)


class HostnameTestCase(base.BaseTestCase):
    def setUp(self):
        super(HostnameTestCase, self).setUp()

        self.test_instance = types.Hostname()

    def test_validate_correct_value(self):
        self.assertTrue(self.test_instance.validate("ns1.ra.restalchemy.com"))
        self.assertTrue(self.test_instance.validate("ns1.55.restalchemy.com"))
        self.assertTrue(self.test_instance.validate("n_s1.55.restalchemy.com"))
        self.assertTrue(self.test_instance.validate("n-1.55.restalchemy.com"))
        self.assertTrue(self.test_instance.validate("restalchemy.com"))

    def test_validate_incorrect_value(self):
        self.assertFalse(self.test_instance.validate("x.y.z"))
        self.assertFalse(self.test_instance.validate("restalchemy.com."))
        self.assertFalse(self.test_instance.validate("restalchemy.com.55"))
        self.assertFalse(self.test_instance.validate("-1.55.restalchemy.com"))
        self.assertFalse(self.test_instance.validate("_s1.55.restalchemy.com"))
        self.assertFalse(self.test_instance.validate(".restalchemy.com"))
        self.assertFalse(self.test_instance.validate("xx.москва.рф"))
        self.assertFalse(self.test_instance.validate("москва.рф"))


class UrlTestCase(base.BaseTestCase):
    def setUp(self):
        super(UrlTestCase, self).setUp()

        self.test_instance = types.Url()

    def test_validate_correct_value(self):
        self.assertTrue(self.test_instance.validate("http://www.gmail.com"))
        self.assertTrue(self.test_instance.validate("https://www.gmail.com/test"))

    def test_validate_incorrect_value(self):
        self.assertFalse(self.test_instance.validate("x.y.z"))
        self.assertFalse(self.test_instance.validate(532))
        self.assertFalse(self.test_instance.validate("google.com.55"))


class SoftSchemeDictTestCase(base.BaseTestCase):
    def setUp(self):
        super(SoftSchemeDictTestCase, self).setUp()

        self.scheme = {
            "sub_str": types.String(),
            "sub_int": types.Integer(),
        }
        self.correct_simple_data = {
            "sub_str": "substr",
            "sub_int": 7,
        }
        self.test_instance = types.SoftSchemeDict(scheme=self.scheme)

    def test_from_simple_type(self):
        self._test_from_simple_type(
            self.test_instance, self.scheme, self.correct_simple_data
        )

    def test_from_simple_type_keys_missing(self):
        data = self.correct_simple_data.copy()
        del data[next(iter(data.keys()))]

        self._test_from_simple_type(self.test_instance, self.scheme, data)

    def test_from_simple_type_empty_data(self):
        self._test_from_simple_type(self.test_instance, self.scheme, {})

    def test_from_simple_type_extra_keys(self):
        data = self.correct_simple_data.copy()
        data["some_key"] = object()

        self.assertRaises(
            KeyError,
            self._test_from_simple_type,
            self.test_instance,
            self.scheme,
            data,
        )

    def test_validate_correct_value(self):
        value = self.test_instance.from_simple_type(self.correct_simple_data)

        self.assertTrue(self._test_validate(self.test_instance, self.scheme, value))

    def test_validate_incorrect_value(self):
        value = self.test_instance.from_simple_type(self.correct_simple_data)
        value[next(iter(value.keys()))] = object()

        self.assertFalse(self._test_validate(self.test_instance, self.scheme, value))

    def test_validate_keys_missing(self):
        data = self.correct_simple_data.copy()
        del data[next(iter(data.keys()))]
        value = self.test_instance.from_simple_type(data)

        self.assertTrue(self._test_validate(self.test_instance, self.scheme, value))

    def test_validate_soft_scheme_dict_empty(self):
        value = self.test_instance.from_simple_type({})

        self.assertTrue(self._test_validate(self.test_instance, self.scheme, value))

    def test_validate_wrong_key(self):
        value = self.test_instance.from_simple_type(self.correct_simple_data)
        first_key = next(iter(value.keys()))
        value[first_key + "something"] = value.pop(first_key)

        self.assertFalse(self._test_validate(self.test_instance, self.scheme, value))

    def _test_from_simple_type(self, typ_obj, scheme, data):
        arg_data = data.copy()

        with mock.patch.object(
            types.Dict, "from_simple_type", return_value=arg_data
        ) as dict_from_simple_type:
            val = typ_obj.from_simple_type(arg_data)

        self.assertEqual(data, arg_data)
        dict_from_simple_type.assert_called_once_with(arg_data)
        self.assertEqual(data, val)
        for key, data_item in data.items():
            self.assertIsInstance(val[key], type(data_item))

    def _test_validate(self, typ_obj, scheme, value):
        arg_value = value.copy()

        with mock.patch.object(
            types.Dict, "validate", return_value=True
        ) as dict_validate:
            ret = typ_obj.validate(arg_value)

        self.assertEqual(value, arg_value)
        dict_validate.assert_called_once_with(arg_value)

        return ret


class SchemaDictTestCase(base.BaseTestCase):
    def test_from_simple_type(self):
        scheme = {
            "sub_str": types.String(),
            "sub_int": types.Integer(),
        }

        test_instance = types.SchemeDict(scheme=scheme)
        self.assertEqual(
            test_instance.from_simple_type({}),
            {
                "sub_str": None,
                "sub_int": None,
            },
        )

    def test_to_simple_type(self):
        scheme = {
            "sub_str": types.String(),
            "sub_int": types.Integer(),
        }

        test_instance = types.SchemeDict(scheme=scheme)
        self.assertEqual(
            test_instance.to_simple_type(
                {
                    "sub_str": None,
                    "sub_int": None,
                }
            ),
            {
                "sub_str": None,
                "sub_int": None,
            },
        )

    def test_validate_wrong(self):
        scheme = {
            "sub_str": types.String(),
            "sub_int": types.Integer(),
        }

        test_instance = types.SchemeDict(scheme=scheme)
        value = test_instance.from_simple_type({})
        self.assertFalse(test_instance.validate(value))

    def test_validate_correct_value_allow_none(self):
        scheme = {
            "sub_str": types.AllowNone(types.String()),
            "sub_int": types.AllowNone(types.Integer()),
        }

        test_instance = types.SchemeDict(scheme=scheme)
        value = test_instance.from_simple_type({})
        self.assertTrue(test_instance.validate(value))

    def test_validate_with_correct_value(self):
        scheme = {
            "sub_str": types.String(),
            "sub_int": types.Integer(),
        }

        test_instance = types.SchemeDict(scheme=scheme)
        value = test_instance.from_simple_type(
            {
                "sub_int": 1,
                "sub_str": "Test",
            }
        )
        self.assertTrue(test_instance.validate(value))

    def test_validate_with_incorrect_value(self):
        scheme = {
            "sub_str": types.String(),
            "sub_int": types.Integer(),
        }

        test_instance = types.SchemeDict(scheme=scheme)
        value = test_instance.from_simple_type(
            {
                "sub_int": "Test",
                "sub_str": "Test",
            }
        )
        self.assertFalse(test_instance.validate(value))

    def test_validate_with_wrong_key(self):
        scheme = {
            "sub_str": types.String(),
            "sub_int": types.Integer(),
        }

        test_instance = types.SchemeDict(scheme=scheme)

        self.assertFalse(
            test_instance.validate(
                {"sub_int": 1, "sub_str": "Test", "sub_wrong": "wrong_key"}
            )
        )

    def test_validate_with_wrong_key_after_from_simple_type(self):
        scheme = {
            "sub_str": types.String(),
            "sub_int": types.Integer(),
        }

        test_instance = types.SchemeDict(scheme=scheme)
        value = test_instance.from_simple_type(
            {
                "sub_int": 1,
                "sub_str": "Test",
                "sub_wrong": "wrong_key",
            }
        )
        self.assertTrue(test_instance.validate(value))

    def test_validate_with_key_missing_after_from_simple_type(self):
        scheme = {
            "sub_str": types.String(),
            "sub_int": types.Integer(),
        }

        test_instance = types.SchemeDict(scheme=scheme)
        value = test_instance.from_simple_type(
            {
                "sub_int": 1,
            }
        )
        self.assertFalse(test_instance.validate(value))


class TimeDeltaTestCase(base.BaseTestCase):
    def setUp(self):
        super().setUp()
        self.test_instance = types.TimeDelta(-10.0, 50.0)

    def test_validate_correct_value(self):
        self.assertTrue(self.test_instance.validate(datetime.timedelta(seconds=10.0)))

    def test_validate_correct_max_value(self):
        self.assertTrue(self.test_instance.validate(datetime.timedelta(seconds=50)))

    def test_validate_correct_min_value(self):
        self.assertTrue(self.test_instance.validate(datetime.timedelta(seconds=-10)))

    def test_validate_incorrect_value(self):
        self.assertFalse(self.test_instance.validate("TEST"))

    def test_validate_incorrect_max_value(self):
        self.assertFalse(self.test_instance.validate(datetime.timedelta(seconds=50.1)))

    def test_validate_incorrect_min_value(self):
        self.assertFalse(self.test_instance.validate(datetime.timedelta(seconds=-10.1)))

    def test_validate_sys_max_value(self):
        test_instance = types.TimeDelta()

        self.assertTrue(
            test_instance.validate(datetime.timedelta(seconds=types.TIMEDELTA_INFINITY))
        )

    def test_validate_sys_min_value(self):
        test_instance = types.TimeDelta()

        self.assertTrue(
            test_instance.validate(
                datetime.timedelta(seconds=-types.TIMEDELTA_INFINITY)
            )
        )



# Copyright 2017 Eugene Frolov <eugene@frolov.net.ru>
#
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import collections
import mock

from mysql.connector import errors
from restalchemy.dm import filters as dm_filters
from restalchemy.dm import models
from restalchemy.dm import properties
from restalchemy.dm import types
from restalchemy.storage.sql.dialect import exceptions as dialect_exc
from restalchemy.storage.sql.dialect import mysql
from restalchemy.storage.sql import tables
from restalchemy.tests.unit import base
from restalchemy.tests import fixtures


class BaseModel(models.ModelWithUUID):
    __tablename__ = "FAKE_TABLE"

    field_int = properties.property(types.Integer())
    field_str = properties.property(types.String())
    field_bool = properties.property(types.Boolean())


class MultipleIdModel(BaseModel):
    tenant_id = properties.property(types.String(), id_property=True)

    @classmethod
    def get_id_property(cls):
        return {"tenant_id": cls.properties["tenant_id"]}


FAKE_TABLE = tables.SQLTable(
    engine=None, table_name=BaseModel.__tablename__, model=BaseModel
)

EXTENDED_FAKE_TABLE = tables.SQLTable(
    engine=None,
    table_name=MultipleIdModel.__tablename__,
    model=MultipleIdModel,
)


FAKE_VALUES = [True, 111, "field2", "uuid"]
FAKE_PK_VALUES = ["uuid"]
EXTENDED_FAKE_VALUES = [True, 111, "field2", "tenant_id", "uuid"]
EXTENDED_FAKE_PK_VALUES = ["uuid", "tenant_id"]


class AbstractDialectCommandTestMixin(object):
    @mock.patch(
        "restalchemy.storage.sql.dialect.base.AbstractDialectCommand.execute",
        side_effect=errors.DatabaseError("deadlock", errno=1213, sqlstate=1213),
    )
    def test_execute_when_errno_1213(self, command_mock):
        with self.assertRaises(dialect_exc.DeadLock) as ctx:
            self.target.execute()
        self.assertEqual(1213, ctx.exception.code)
        self.assertEqual("deadlock", str(ctx.exception))

    @mock.patch(
        "restalchemy.storage.sql.dialect.base.AbstractDialectCommand.execute",
        side_effect=errors.DatabaseError("conflict", errno=1062, sqlstate=1062),
    )
    def test_execute_when_errno_1062(self, command_mock):
        with self.assertRaises(dialect_exc.Conflict) as ctx:
            self.target.execute()
        self.assertEqual(1062, ctx.exception.code)
        self.assertEqual("conflict", str(ctx.exception))

    @mock.patch(
        "restalchemy.storage.sql.dialect.base.AbstractDialectCommand.execute",
        side_effect=errors.DatabaseError("access denied", errno=1045),
    )
    def test_execute_when_errno_1045(self, command_mock):
        with self.assertRaises(errors.DatabaseError) as ctx:
            self.target.execute()
        self.assertEqual(1045, ctx.exception.errno)
        self.assertEqual("access denied", ctx.exception.msg)


class MySQLInsertTestCase(base.BaseTestCase, AbstractDialectCommandTestMixin):
    def setUp(self):
        self.target = mysql.MySQLInsert(
            FAKE_TABLE,
            FAKE_VALUES,
            session=fixtures.SessionFixture(),
        )

    def test_statement(self):
        self.assertEqual(
            self.target.get_statement(),
            "INSERT INTO `FAKE_TABLE` (`field_bool`, `field_int`, "
            "`field_str`, `uuid`) VALUES (%s, %s, %s, %s)",
        )


class MySQLUpdateTestCase(base.BaseTestCase, AbstractDialectCommandTestMixin):
    def setUp(self):
        TABLE = FAKE_TABLE
        self.target = mysql.MySQLUpdate(
            TABLE,
            FAKE_PK_VALUES,
            FAKE_VALUES,
            session=fixtures.SessionFixture(),
        )

    def test_statement(self):
        self.assertEqual(
            self.target.get_statement(),
            "UPDATE `FAKE_TABLE` SET `field_bool` = %s, `field_int` = %s, "
            "`field_str` = %s WHERE `uuid` = %s",
        )


class MySQLUpdateMultipleIdTestCase(base.BaseTestCase, AbstractDialectCommandTestMixin):
    def setUp(self):
        TABLE = EXTENDED_FAKE_TABLE
        self.target = mysql.MySQLUpdate(
            TABLE,
            EXTENDED_FAKE_PK_VALUES,
            EXTENDED_FAKE_VALUES,
            session=fixtures.SessionFixture(),
        )

    def test_statement(self):
        self.assertEqual(
            self.target.get_statement(),
            "UPDATE `FAKE_TABLE` SET `field_bool` = %s, `field_int` = %s, "
            "`field_str` = %s WHERE `tenant_id` = %s AND `uuid` = %s",
        )


class MySQLDeleteTestCase(base.BaseTestCase, AbstractDialectCommandTestMixin):
    def setUp(self):
        TABLE = FAKE_TABLE

        self.target = mysql.MySQLDelete(
            TABLE,
            FAKE_PK_VALUES,
            session=fixtures.SessionFixture(),
        )

    def test_statement(self):
        self.assertEqual(
            self.target.get_statement(),
            "DELETE FROM `FAKE_TABLE` WHERE `uuid` = %s",
        )


class MySQLDeleteMultipleIdTestCase(base.BaseTestCase, AbstractDialectCommandTestMixin):
    def setUp(self):
        TABLE = EXTENDED_FAKE_TABLE

        self.target = mysql.MySQLDelete(
            TABLE,
            EXTENDED_FAKE_VALUES,
            session=fixtures.SessionFixture(),
        )

    def test_statement(self):
        self.assertEqual(
            self.target.get_statement(),
            "DELETE FROM `FAKE_TABLE` WHERE `tenant_id` = %s AND `uuid` = %s",
        )


class MySQLSelectTestCase(base.BaseTestCase):
    def setUp(self):
        self._TABLE = FAKE_TABLE

    def test_statement_OR(self):
        session = mock.Mock()
        ord_filter = collections.OrderedDict()
        for k, v in sorted(zip(self._TABLE.get_column_names(session), FAKE_VALUES)):
            ord_filter[k] = dm_filters.EQ(v)
        FAKE_EQ_VALUES = dm_filters.OR(ord_filter)
        target = mysql.MySQLSelect(
            self._TABLE,
            filters=FAKE_EQ_VALUES,
            session=fixtures.SessionFixture(),
        )

        result = target.get_statement()

        self.assertEqual(
            "SELECT `field_bool`, `field_int`, `field_str`, `uuid` "
            "FROM `FAKE_TABLE` WHERE (`field_bool` = %s OR "
            "`field_int` = %s OR `field_str` = %s OR `uuid` = %s)",
            result,
        )
        self.assertEqual(FAKE_VALUES, target.get_values())

    def test_statement_recursive_OR(self):
        FAKE_EQ_VALUES = dm_filters.OR(
            dm_filters.AND(
                {"field_bool": dm_filters.EQ(True)},
                {"field_int": dm_filters.EQ(111)},
            ),
            dm_filters.AND(
                {"field_str": dm_filters.EQ("field2")},
                {"uuid": dm_filters.EQ("uuid")},
            ),
        )
        target = mysql.MySQLSelect(
            self._TABLE,
            filters=FAKE_EQ_VALUES,
            session=fixtures.SessionFixture(),
        )

        result = target.get_statement()

        self.assertEqual(
            "SELECT `field_bool`, `field_int`, `field_str`, `uuid` "
            "FROM `FAKE_TABLE` WHERE ((`field_bool` = %s AND "
            "`field_int` = %s) OR (`field_str` = %s AND `uuid` = %s))",
            result,
        )
        self.assertEqual(FAKE_VALUES, target.get_values())

    def test_statement_EQ(self):
        session = mock.Mock()
        FAKE_EQ_VALUES = dm_filters.AND(
            *[
                {k: dm_filters.EQ(v)}
                for k, v in sorted(
                    zip(self._TABLE.get_column_names(session), FAKE_VALUES)
                )
            ]
        )
        target = mysql.MySQLSelect(
            self._TABLE,
            filters=FAKE_EQ_VALUES,
            session=fixtures.SessionFixture(),
        )

        result = target.get_statement()

        self.assertEqual(
            "SELECT `field_bool`, `field_int`, `field_str`, `uuid` "
            "FROM `FAKE_TABLE` WHERE (`field_bool` = %s AND "
            "`field_int` = %s AND `field_str` = %s AND `uuid` = %s)",
            result,
        )
        self.assertEqual(FAKE_VALUES, target.get_values())

    def test_statement_NE(self):
        session = mock.Mock()
        FAKE_NE_VALUES = dm_filters.AND(
            *[
                {k: dm_filters.NE(v)}
                for k, v in sorted(
                    zip(self._TABLE.get_column_names(session), FAKE_VALUES)
                )
            ]
        )
        target = mysql.MySQLSelect(
            self._TABLE,
            filters=FAKE_NE_VALUES,
            session=fixtures.SessionFixture(),
        )

        result = target.get_statement()

        self.assertEqual(
            "SELECT `field_bool`, `field_int`, `field_str`, `uuid` "
            "FROM `FAKE_TABLE` WHERE (`field_bool` <> %s AND "
            "`field_int` <> %s AND `field_str` <> %s AND `uuid` <> %s)",
            result,
        )
        self.assertEqual(FAKE_VALUES, target.get_values())

    def test_statement_GT(self):
        session = mock.Mock()
        FAKE_GT_VALUES = dm_filters.AND(
            *[
                {k: dm_filters.GT(v)}
                for k, v in sorted(
                    zip(self._TABLE.get_column_names(session), FAKE_VALUES)
                )
            ]
        )
        target = mysql.MySQLSelect(
            self._TABLE,
            filters=FAKE_GT_VALUES,
            session=fixtures.SessionFixture(),
        )

        result = target.get_statement()

        self.assertEqual(
            "SELECT `field_bool`, `field_int`, `field_str`, `uuid` "
            "FROM `FAKE_TABLE` WHERE (`field_bool` > %s AND "
            "`field_int` > %s AND `field_str` > %s AND `uuid` > %s)",
            result,
        )
        self.assertEqual(FAKE_VALUES, target.get_values())

    def test_statement_GE(self):
        session = mock.Mock()
        FAKE_GE_VALUES = dm_filters.AND(
            *[
                {k: dm_filters.GE(v)}
                for k, v in sorted(
                    zip(self._TABLE.get_column_names(session), FAKE_VALUES)
                )
            ]
        )
        target = mysql.MySQLSelect(
            self._TABLE,
            filters=FAKE_GE_VALUES,
            session=fixtures.SessionFixture(),
        )

        result = target.get_statement()

        self.assertEqual(
            "SELECT `field_bool`, `field_int`, `field_str`, `uuid` "
            "FROM `FAKE_TABLE` WHERE (`field_bool` >= %s AND "
            "`field_int` >= %s AND `field_str` >= %s AND `uuid` >= %s)",
            result,
        )
        self.assertEqual(FAKE_VALUES, target.get_values())

    def test_statement_LT(self):
        session = mock.Mock()
        FAKE_LT_VALUES = dm_filters.AND(
            *[
                {k: dm_filters.LT(v)}
                for k, v in sorted(
                    zip(self._TABLE.get_column_names(session), FAKE_VALUES)
                )
            ]
        )
        target = mysql.MySQLSelect(
            self._TABLE,
            filters=FAKE_LT_VALUES,
            session=fixtures.SessionFixture(),
        )

        result = target.get_statement()

        self.assertEqual(
            "SELECT `field_bool`, `field_int`, `field_str`, `uuid` "
            "FROM `FAKE_TABLE` WHERE (`field_bool` < %s AND "
            "`field_int` < %s AND `field_str` < %s AND `uuid` < %s)",
            result,
        )
        self.assertEqual(FAKE_VALUES, target.get_values())

    def test_statement_LE(self):
        session = mock.Mock()
        FAKE_LE_VALUES = dm_filters.AND(
            *[
                {k: dm_filters.LE(v)}
                for k, v in sorted(
                    zip(self._TABLE.get_column_names(session), FAKE_VALUES)
                )
            ]
        )
        target = mysql.MySQLSelect(
            self._TABLE,
            filters=FAKE_LE_VALUES,
            session=fixtures.SessionFixture(),
        )

        result = target.get_statement()

        self.assertEqual(
            "SELECT `field_bool`, `field_int`, `field_str`, `uuid` "
            "FROM `FAKE_TABLE` WHERE (`field_bool` <= %s AND "
            "`field_int` <= %s AND `field_str` <= %s AND `uuid` <= %s)",
            result,
        )
        self.assertEqual(FAKE_VALUES, target.get_values())

    def test_statement_limit_with_where_clause(self):
        session = mock.Mock()
        FAKE_LE_VALUES = dm_filters.AND(
            *[
                {k: dm_filters.LE(v)}
                for k, v in sorted(
                    zip(self._TABLE.get_column_names(session), FAKE_VALUES)
                )
            ]
        )
        target = mysql.MySQLSelect(
            self._TABLE,
            filters=FAKE_LE_VALUES,
            limit=2,
            session=fixtures.SessionFixture(),
        )

        result = target.get_statement()

        self.assertEqual(
            "SELECT `field_bool`, `field_int`, `field_str`, `uuid` "
            "FROM `FAKE_TABLE` WHERE (`field_bool` <= %s AND "
            "`field_int` <= %s AND `field_str` <= %s AND `uuid` <= %s) "
            "LIMIT 2",
            result,
        )
        self.assertEqual(FAKE_VALUES, target.get_values())

    def test_statement_locked_with_where_clause(self):
        session = mock.Mock()
        FAKE_LE_VALUES = dm_filters.AND(
            *[
                {k: dm_filters.LE(v)}
                for k, v in sorted(
                    zip(self._TABLE.get_column_names(session), FAKE_VALUES)
                )
            ]
        )
        target = mysql.MySQLSelect(
            self._TABLE,
            filters=FAKE_LE_VALUES,
            locked=True,
            session=fixtures.SessionFixture(),
        )

        result = target.get_statement()

        self.assertEqual(
            "SELECT `field_bool`, `field_int`, `field_str`, `uuid` "
            "FROM `FAKE_TABLE` WHERE (`field_bool` <= %s AND "
            "`field_int` <= %s AND `field_str` <= %s AND `uuid` <= %s) "
            "FOR UPDATE",
            result,
        )
        self.assertEqual(FAKE_VALUES, target.get_values())

    def test_statement_order_by_with_where_clause(self):
        session = mock.Mock()
        orders = collections.OrderedDict()
        orders["field_str"] = ""
        orders["field_bool"] = "desc"
        FAKE_LE_VALUES = dm_filters.AND(
            *[
                {k: dm_filters.LE(v)}
                for k, v in sorted(
                    zip(self._TABLE.get_column_names(session), FAKE_VALUES)
                )
            ]
        )
        target = mysql.MySQLSelect(
            self._TABLE,
            filters=FAKE_LE_VALUES,
            order_by=orders,
            session=fixtures.SessionFixture(),
        )

        result = target.get_statement()

        self.assertEqual(
            "SELECT `field_bool`, `field_int`, `field_str`, `uuid` "
            "FROM `FAKE_TABLE` WHERE (`field_bool` <= %s AND "
            "`field_int` <= %s AND `field_str` <= %s AND `uuid` <= %s) "
            "ORDER BY `field_str` ASC, `field_bool` DESC",
            result,
        )
        self.assertEqual(FAKE_VALUES, target.get_values())

    def test_statement_order_by_without_where_clause(self):
        orders = collections.OrderedDict()
        orders["field_str"] = ""
        orders["field_bool"] = "desc"
        target = mysql.MySQLSelect(
            self._TABLE,
            order_by=orders,
            session=fixtures.SessionFixture(),
        )

        result = target.get_statement()

        self.assertEqual(
            "SELECT `field_bool`, `field_int`, `field_str`, `uuid` "
            "FROM `FAKE_TABLE` "
            "ORDER BY `field_str` ASC, `field_bool` DESC",
            result,
        )
        self.assertEqual([], target.get_values())

    def test_statement_order_by_false_order(self):
        session = mock.Mock()
        FAKE_LE_VALUES = dm_filters.AND(
            *[
                {k: dm_filters.LE(v)}
                for k, v in sorted(
                    zip(self._TABLE.get_column_names(session), FAKE_VALUES)
                )
            ]
        )
        target = mysql.MySQLSelect(
            self._TABLE,
            filters=FAKE_LE_VALUES,
            order_by={"field_str": "FALSE"},
            session=fixtures.SessionFixture(),
        )
        self.assertRaises(ValueError, target.get_statement)


class MySQLCustomSelectTestCase(base.BaseTestCase):
    def setUp(self):
        self._TABLE = FAKE_TABLE

    def test_custom_where_condition(self):
        FAKE_WHERE_CONDITION = "NOT (`field_int` => %s AND `field_str` = %s)"
        FAKE_WHERE_VALUES = [1, "2"]
        target = mysql.MySQLCustomSelect(
            self._TABLE,
            FAKE_WHERE_CONDITION,
            FAKE_WHERE_VALUES,
            session=fixtures.SessionFixture(),
        )

        result = target.get_statement()

        self.assertEqual(
            "SELECT `field_bool`, `field_int`, `field_str`, `uuid` "
            "FROM `FAKE_TABLE` WHERE "
            "NOT (`field_int` => %s AND `field_str` = %s)",
            result,
        )

    def test_custom_where_condition_with_limit(self):
        FAKE_WHERE_CONDITION = "NOT (`field_int` => %s AND `field_str` = %s)"
        FAKE_WHERE_VALUES = [1, "2"]
        target = mysql.MySQLCustomSelect(
            self._TABLE,
            FAKE_WHERE_CONDITION,
            FAKE_WHERE_VALUES,
            limit=2,
            session=fixtures.SessionFixture(),
        )

        result = target.get_statement()

        self.assertEqual(
            "SELECT `field_bool`, `field_int`, `field_str`, `uuid` "
            "FROM `FAKE_TABLE` WHERE "
            "NOT (`field_int` => %s AND `field_str` = %s) LIMIT 2",
            result,
        )

    def test_custom_where_condition_with_locked(self):
        FAKE_WHERE_CONDITION = "NOT (`field_int` => %s AND `field_str` = %s)"
        FAKE_WHERE_VALUES = [1, "2"]
        target = mysql.MySQLCustomSelect(
            self._TABLE,
            FAKE_WHERE_CONDITION,
            FAKE_WHERE_VALUES,
            locked=True,
            session=fixtures.SessionFixture(),
        )

        result = target.get_statement()

        self.assertEqual(
            "SELECT `field_bool`, `field_int`, `field_str`, `uuid` "
            "FROM `FAKE_TABLE` WHERE "
            "NOT (`field_int` => %s AND `field_str` = %s) FOR UPDATE",
            result,
        )

    def test_custom_where_condition_with_order_by(self):
        FAKE_WHERE_CONDITION = "NOT (`field_int` => %s AND `field_str` = %s)"
        FAKE_WHERE_VALUES = [1, "2"]
        target = mysql.MySQLCustomSelect(
            self._TABLE,
            FAKE_WHERE_CONDITION,
            FAKE_WHERE_VALUES,
            order_by={"field_str": ""},
            session=fixtures.SessionFixture(),
        )

        result = target.get_statement()

        self.assertEqual(
            "SELECT `field_bool`, `field_int`, `field_str`, `uuid` "
            "FROM `FAKE_TABLE` WHERE "
            "NOT (`field_int` => %s AND `field_str` = %s) "
            "ORDER BY `field_str` ASC",
            result,
        )


class MySQLCountTestCase(base.BaseTestCase):
    def setUp(self):
        self._TABLE = FAKE_TABLE

    def test_statement(self):
        target = mysql.MySQLCount(
            self._TABLE,
            session=fixtures.SessionFixture(),
        )

        self.assertEqual(
            target.get_statement(),
            "SELECT COUNT(*) as count FROM `FAKE_TABLE`",
        )

    def test_statement_where(self):
        session = mock.Mock()
        FAKE_EQ_VALUES = dm_filters.AND(
            *[
                {k: dm_filters.EQ(v)}
                for k, v in sorted(
                    zip(self._TABLE.get_column_names(session), FAKE_VALUES)
                )
            ]
        )
        target = mysql.MySQLCount(
            self._TABLE,
            filters=FAKE_EQ_VALUES,
            session=fixtures.SessionFixture(),
        )

        self.assertEqual(
            (
                "SELECT COUNT(*) as count FROM `FAKE_TABLE` "
                "WHERE (`field_bool` = %s "
                "AND `field_int` = %s AND `field_str` = %s AND `uuid` = %s)"
            ),
            target.get_statement(),
        )



#    Copyright 2014 Eugene Frolov <eugene@frolov.net.ru>
#    Copyright 2021 Eugene Frolov.
#
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from restalchemy.api import constants
from restalchemy.common import exceptions as exc


class ActionHandler(object):
    def __init__(self, get=None, post=None, put=None):
        self._get = get
        self._post = post
        self._put = put
        self._method_actions_map = {
            get: constants.ACTION_GET,
            post: constants.ACTION_POST,
            put: constants.ACTION_PUT,
        }
        if get:
            self._name = get.__name__
            self._method = constants.GET
        elif post:
            self._name = post.__name__
            self._method = constants.POST
        elif put:
            self._name = put.__name__
            self._method = constants.PUT

    def get(self, fn):
        self._get = fn
        return self

    @property
    def name(self):
        return self._name

    @property
    def method(self):
        return self._method

    def post(self, fn):
        self._post = fn
        return self

    def put(self, fn):
        self._put = fn
        return self

    def do(self, fn, controller, *args, **kwargs):
        if fn:
            api_context = controller.request.api_context
            api_context.set_active_method(self._method_actions_map[fn])
            result = fn(self=controller, *args, **kwargs)
            return controller.process_result(result=result)
        else:
            raise exc.NotImplementedError()

    def do_get(self, *args, **kwargs):
        return self.do(self._get, *args, **kwargs)

    def do_post(self, *args, **kwargs):
        return self.do(self._post, *args, **kwargs)

    def do_put(self, *args, **kwargs):
        return self.do(self._put, *args, **kwargs)


def get(fn):
    return ActionHandler(get=fn)


def post(fn):
    return ActionHandler(post=fn)


def put(fn):
    return ActionHandler(put=fn)



#    Copyright 2020 Eugene Frolov.
#
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
import sys

import pytest

from restalchemy.storage import base as sbase
from restalchemy.storage import exceptions
from restalchemy.tests.unit import base


class TestErrorCatcherTestCase(base.BaseTestCase):
    @sbase.error_catcher
    def my_func(self, *args, **kwargs):
        if args[0] == "test_arg" and kwargs.get("kwarg0") == "test_kwarg":
            raise RuntimeError(
                "Some Error!!!",
            )
        elif args[0] == "test_arg" and kwargs.get("kwarg0") == "RA_EXCEPTION":
            raise exceptions.RecordNotFound(model="FAKE", filters="FAKE")

    def test_catcher(self):
        self.assertRaises(
            exceptions.UnknownStorageException,
            self.my_func,
            "test_arg",
            kwarg0="test_kwarg",
        )

    def test_catcher_ra_exception(self):
        self.assertRaises(
            exceptions.RecordNotFound,
            self.my_func,
            "test_arg",
            kwarg0="RA_EXCEPTION",
        )

    @pytest.mark.skipif(sys.version_info > (3, 6), reason="python > 3.6")
    def test_catcher_message_error_lt_36(self):
        try:
            self.my_func("test_arg", kwarg0="test_kwarg")
            raise AssertionError("The exception is't raised")
        except exceptions.UnknownStorageException as e:
            self.assertEqual(
                str(e),
                "Unknown storage exception: RuntimeError('Some Error!!!',)",
            )

    @pytest.mark.skipif(sys.version_info < (3, 7), reason="python < 3.7")
    def test_catcher_message_error_gt_36(self):
        try:
            self.my_func("test_arg", kwarg0="test_kwarg")
            raise AssertionError("The exception is't raised")
        except exceptions.UnknownStorageException as e:
            self.assertEqual(
                str(e),
                "Unknown storage exception: RuntimeError('Some Error!!!')",
            )



# Copyright 2016 Eugene Frolov <eugene@frolov.net.ru>
#
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from restalchemy.storage.sql import migrations


class MigrationStep(migrations.AbstractMigrationStep):
    def __init__(self):
        self._depends = ["0001-rest-service-tables-migration-e31a12.py"]

    @property
    def migration_id(self):
        return "c17a6066-fd95-4fad-84ed-c8777dff9a08"

    def upgrade(self, session):
        expressions = [
            """
                INSERT INTO vms (uuid, name, state)
                VALUES ('00000000-0000-0000-0000-000000000001', 'vm1', 'on'),
                       ('00000000-0000-0000-0000-000000000002', 'vm2', 'off');
            """
        ]

        for expression in expressions:
            session.execute(expression)

    def downgrade(self, session):
        expressions = [
            """
                DELETE from tags;
            """,
            """
                DELETE FROM ports;
            """,
            """
                DELETE FROM vms;
            """,
        ]

        for expression in expressions:
            session.execute(expression)


migration_step = MigrationStep()



# Copyright 2019 Eugene Frolov
#
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import mock
from mock import patch

from mysql.connector import errors
from restalchemy.storage import exceptions as exc
from restalchemy.storage.sql import sessions
from restalchemy.tests.unit import base


class TestLocalThreadStorage(base.BaseTestCase):
    def setUp(self):
        super(TestLocalThreadStorage, self).setUp()
        self._storage = sessions.SessionThreadStorage()

    def tearDown(self):
        super(TestLocalThreadStorage, self).tearDown()
        self._storage.remove_session()
        del self._storage

    def test_store_session(self):
        session = mock.Mock()

        self.assertEqual(self._storage.store_session(session), session)
        self.assertEqual(self._storage.get_session(), session)

    def test_store_session_conflict_exception(self):
        session = mock.Mock()

        with patch.object(self._storage, "get_session", return_value=session):
            self.assertRaises(
                sessions.SessionConflict, self._storage.store_session, session
            )

    def test_store_session_conflict_exception_with_two_instance(self):
        instance_one = sessions.SessionThreadStorage()
        instance_two = sessions.SessionThreadStorage()
        session1 = mock.Mock()
        session2 = mock.Mock()

        instance_one.store_session(session1)

        with self.assertRaises(sessions.SessionConflict):
            instance_two.store_session(session2)

    def test_get_session_not_found_exception(self):

        self.assertRaises(sessions.SessionNotFound, self._storage.get_session)

    def test_get_session(self):
        session = mock.Mock()
        self._storage._storage.session = session

        self.assertEqual(self._storage.get_session(), session)

    def test_pop_session(self):
        session = mock.Mock()
        self._storage._storage.session = session

        result = self._storage.pop_session()

        self.assertEqual(result, session)
        self.assertIsNone(self._storage._storage.session)

    def test_execute_when_deadlock_raises(self):
        mock_execute = mock.Mock()
        mock_execute.execute.side_effect = errors.DatabaseError(
            "deadlock", errno=1213, sqlstate=1213
        )
        mock_cursor = mock.Mock()
        mock_cursor.cursor.return_value = mock_execute
        mock_engine = mock.Mock()
        mock_engine.get_connection.return_value = mock_cursor
        session = sessions.MySQLSession(mock_engine)
        with self.assertRaises(exc.DeadLock) as ctx:
            session.execute("update foo set bar = 'baz'")
        self.assertEqual(
            "Deadlock found when trying to get lock. Original message: deadlock",
            str(ctx.exception),
        )

    def test_execute_when_unknown_exc_raises(self):
        mock_execute = mock.Mock()
        mock_execute.execute.side_effect = errors.DatabaseError(
            "error", errno=1062, sqlstate=1062
        )
        mock_cursor = mock.Mock()
        mock_cursor.cursor.return_value = mock_execute
        mock_engine = mock.Mock()
        mock_engine.get_connection.return_value = mock_cursor
        session = sessions.MySQLSession(mock_engine)
        with self.assertRaises(errors.DatabaseError) as ctx:
            session.execute("update foo set bar = 'baz'")
        self.assertEqual("1062 (1062): error", str(ctx.exception))



# Copyright 2016 Eugene Frolov <eugene@frolov.net.ru>
#
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import abc
import logging
import os
import re
import sys
import uuid

from restalchemy.common import contexts
from restalchemy.dm import filters
from restalchemy.dm import models
from restalchemy.dm import properties
from restalchemy.dm import types
from restalchemy.storage import exceptions
from restalchemy.storage.sql import orm

HEAD_MIGRATION = "HEAD"
MANUAL_MIGRATION = "MANUAL"

RA_MIGRATION_TABLE_NAME = "ra_migrations"
LOG = logging.getLogger(__name__)

DEFAULT_NEW_MIGRATION_NUMBER = "0000"
MIGRATION_NUMBER_LENGTH = len(DEFAULT_NEW_MIGRATION_NUMBER)
OLD_STYLE_NAME_PATTERN = r"[0-9a-z]{6}"
OLD_STYLE = re.compile(OLD_STYLE_NAME_PATTERN)


class HeadMigrationNotFoundException(Exception):
    pass


class DependenciesException(Exception):
    pass


class AbstractMigrationStep(metaclass=abc.ABCMeta):
    @property
    def depends(self):
        return [dep for dep in self._depends if dep]

    @property
    @abc.abstractmethod
    def migration_id(self):
        raise NotImplementedError()

    @property
    def is_manual(self):
        return False

    @property
    def number(self):
        return self._migration_number

    @number.setter
    def number(self, value):
        self._migration_number = value

    @abc.abstractmethod
    def upgrade(self, session):
        raise NotImplementedError()

    @abc.abstractmethod
    def downgrade(self, session):
        raise NotImplementedError()

    @staticmethod
    def _delete_table_if_exists(session, table_name):
        session.execute(f"DROP TABLE IF EXISTS {session.engine.escape(table_name)};")

    @staticmethod
    def _delete_trigger_if_exists(session, trigger_name):
        session.execute(
            f"DROP TRIGGER IF EXISTS {session.engine.escape(trigger_name)};"
        )

    @staticmethod
    def _delete_view_if_exists(session, view_name):
        session.execute(f"DROP VIEW IF EXISTS {session.engine.escape(view_name)};")


class AbstarctMigrationStep(AbstractMigrationStep):
    """Legacy class with typo in name, please migrate to valid class"""

    pass


class MigrationModel(models.ModelWithUUID, orm.SQLStorableMixin):
    __tablename__ = RA_MIGRATION_TABLE_NAME

    applied = properties.property(types.Boolean(), required=True, default=False)


class MigrationStepController(object):
    def __init__(self, migration_step, filename, session):
        self._migration_step = migration_step
        self._filename = filename
        migr_uuid = uuid.UUID(self._migration_step.migration_id)
        try:
            self._migration_model = MigrationModel.objects.get_one(
                filters={"uuid": filters.EQ(migr_uuid)}, session=session
            )
        except exceptions.RecordNotFound:
            self._migration_model = MigrationModel(
                uuid=uuid.UUID(self._migration_step.migration_id)
            )

    def is_applied(self):
        return self._migration_model.applied

    def is_manual(self):
        return self._migration_step.is_manual

    def depends_from(self):
        return self._migration_step.depends

    def apply(self, session, migrations, dry_run=False):
        if self.is_applied():
            LOG.warning("Migration '%s' is already applied", self.name)
            return

        LOG.info(
            "Migration '%s' depends on %r",
            self.name,
            self._migration_step.depends,
        )

        for depend in self._migration_step.depends:
            migrations[depend].apply(session, migrations, dry_run=dry_run)

        if dry_run:
            LOG.info("Dry run upgrade for migration '%s'", self.name)
            return

        self._migration_step.upgrade(session)
        self._migration_model.applied = True
        self._migration_model.save(session=session)

    def rollback(self, session, migrations, dry_run=False):
        if not self.is_applied():
            LOG.warning("Migration '%s' is not applied.", self.name)
            return

        for migration in migrations.values():
            if self._filename in migration.depends_from():
                LOG.info("Migration '%s' dependent %r", self.name, migration.name)
                migration.rollback(session, migrations, dry_run=dry_run)

        if dry_run:
            LOG.info("Dry run downgrade for migration '%s'", self.name)
            return

        self._migration_step.downgrade(session)
        self._migration_model.applied = False
        self._migration_model.save(session=session)

    @property
    def name(self):
        return os.path.splitext(os.path.basename(self._filename))[0]


class MigrationEngine(object):
    FILENAME_HASH_LEN = 6

    def __init__(self, migrations_path):
        self._migrations_path = migrations_path

    def get_file_name(self, part_of_name):
        candidates = []
        for filename in os.listdir(self._migrations_path):
            if part_of_name in filename and filename.endswith(".py"):
                candidates.append(filename)

        candidates_count = len(candidates)

        if candidates_count == 1:
            return candidates[0]

        if candidates_count > 1:
            raise ValueError(
                "Multiple file found for name '%s': %s." % (part_of_name, candidates)
            )

        raise ValueError("Migration file for dependency %s not found" % part_of_name)

    def _calculate_depends(self, depends):
        files = []

        for depend in depends:
            if depend.upper() == HEAD_MIGRATION:
                file_name = self.get_latest_migration()
            else:
                file_name = self.get_file_name(depend)
            files.append(file_name)
        return files

    def _suggest_new_migration_number(self):
        suggestion_number = DEFAULT_NEW_MIGRATION_NUMBER
        migrations = self._load_migrations()
        migration_numbers = set()
        for migration in migrations:
            parts = migration.split("-")
            if len(parts) > 1:
                number_part = parts[0].rstrip(".py")
                if OLD_STYLE.match(number_part):
                    number_part = parts[1].rstrip("py")
                if number_part.isdigit():
                    migration_numbers.add(number_part)

        if not migration_numbers:
            return suggestion_number

        max_number = int(max(migration_numbers, key=int))
        max_len = len(max(migration_numbers, key=len))

        suggestion_number = str(max_number + 1).zfill(max_len)
        return suggestion_number

    def new_migration(self, depends, message, dry_run=False, is_manual=False):
        files = self._calculate_depends(depends)
        depends = '"{}"'.format('", "'.join(files)) if files else ""

        migration_id = str(uuid.uuid4())
        filename_hash = migration_id[: self.FILENAME_HASH_LEN]
        migration_number = (
            MANUAL_MIGRATION if is_manual else self._suggest_new_migration_number()
        )
        message = message.replace(" ", "-")

        first_part_of_message = message.split("-")[0]

        if (
            first_part_of_message.isdigit()
            and len(first_part_of_message) == MIGRATION_NUMBER_LENGTH
        ):
            migration_number = first_part_of_message
            message = message.lstrip(first_part_of_message + "-")

        mfilename = "-".join([migration_number, message, filename_hash]) + ".py"

        mpath = os.path.join(self._migrations_path, mfilename)

        if dry_run:
            LOG.info(
                "Dry run create migration '%s'. File: %s, path: %s",
                message,
                mfilename,
                mpath,
            )
            return

        with open(mpath, "w") as fp_output:
            template_path = os.path.join(
                os.path.dirname(__file__), "migration_templ.tmpl"
            )
            with open(template_path, "r") as fp_input:
                fp_output.write(
                    fp_input.read()
                    % {
                        "migration_id": migration_id,
                        "depends": depends,
                        "is_manual": is_manual,
                    }
                )

        LOG.info("New migration '%s' has been created: %s", mfilename, mpath)

    @staticmethod
    def _init_migration_table(session):
        statement = f"""CREATE TABLE IF NOT EXISTS {RA_MIGRATION_TABLE_NAME} (
            uuid CHAR(36) NOT NULL PRIMARY KEY,
            applied BOOLEAN NOT NULL
        )"""
        session.execute(statement, None)

    def _load_migrations(self):
        migrations = {}
        sys.path.insert(0, self._migrations_path)
        try:
            for filename in os.listdir(self._migrations_path):
                if filename.endswith(".py"):
                    migration = __import__(filename[:-3])
                    if not hasattr(migration, "migration_step"):
                        continue
                    migrations[filename] = migration.migration_step
            return migrations
        finally:
            sys.path.remove(self._migrations_path)

    def _load_migration_controllers(self, session):
        return {
            filename: MigrationStepController(
                migration_step=step,
                filename=filename,
                session=session,
            )
            for filename, step in self._load_migrations().items()
        }

    def apply_migration(self, migration_name, dry_run=False):
        filename = self.get_file_name(migration_name)
        with contexts.Context().session_manager() as session:
            self._init_migration_table(session)
            migrations = self._load_migration_controllers(session)

            migration = migrations[filename]
            if migration.is_applied():
                LOG.warning("Migration '%s' is already applied", migration.name)
            else:
                LOG.info("Applying migration '%s'", migration.name)
                migrations[filename].apply(session, migrations, dry_run=dry_run)

    def rollback_migration(self, migration_name, dry_run=False):
        filename = self.get_file_name(migration_name)
        with contexts.Context().session_manager() as session:
            self._init_migration_table(session)
            migrations = self._load_migration_controllers(session)
            migration = migrations[filename]
            if not migration.is_applied():
                LOG.warning("Migration '%s' is not applied", migration.name)
            else:
                LOG.info("Rolling back migration '%s'", migration.name)
                migrations[filename].rollback(session, migrations, dry_run=dry_run)

    def _calculate_indexes(self):
        indexed_migrations = []
        migrations = self._load_migrations()
        current_migration = self.get_latest_migration()

        indexed_migrations.append(current_migration)

        while True:
            if not migrations[current_migration].depends:
                break
            for d in migrations[current_migration].depends:
                indexed_migrations.append(d)
                current_migration = d

        indexed_migrations.reverse()

        for index, filename in enumerate(indexed_migrations):
            migrations[filename].number = index
        return migrations

    def get_all_migrations(self):
        return {
            filename: {
                "is_manual": m.is_manual,
                "depends": m.depends,
                "uuid": m.migration_id,
                "index": m.number if hasattr(m, "number") else 0,
            }
            for filename, m in self._calculate_indexes().items()
        }

    def get_latest_migration(self):
        migrations = {
            filename: m
            for filename, m in self._load_migrations().items()
            if m.is_manual is False
        }

        for migration in list(migrations.values()):
            for depend in migration._depends:
                if depend in migrations:
                    migrations.pop(depend, None)
        if len(migrations) == 1:
            return migrations.popitem()[0]

        raise HeadMigrationNotFoundException(
            "Head migration for current migrations couldn't be found"
        )

    def validate_auto_migration_dependencies(self, depends):
        depends = self._calculate_depends(depends)

        migrations = self._load_migrations()

        for filename in depends:
            if migrations[filename].is_manual:
                LOG.warning("Manual migration(s) is(are) in dependencies!")
                return False
        return True

    def get_unapplied_migrations(self, session, include_manual=False):
        self._init_migration_table(session)
        migrations = self._load_migration_controllers(session)

        filtered_migrations = {}
        for filename, migration in migrations.items():
            if migration.is_applied() is False:
                if migration.is_manual() is False or include_manual is True:
                    filtered_migrations[filename] = migration
        return filtered_migrations



# Copyright 2022 Eugene Frolov <eugene@frolov.net.ru>
#
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import collections
import posixpath

from restalchemy.api import constants as c
from restalchemy.common import utils
from restalchemy.openapi import constants as oa_c
from restalchemy.openapi import utils as oa_utils


class OpenApiContact(object):
    """Contact Object

    Contact information for the exposed API.

    :param name: The identifying name of the contact person/organization.
    :type name: str
    :param url: The URL pointing to the contact information. MUST be in the
        format of a URL.
    :type url: str
    :param email: The email address of the contact person/organization. MUST
        be in the format of an email address.
    :type email: str
    """

    def __init__(self, name=None, url=None, email=None):
        super(OpenApiContact, self).__init__()
        self._name = name
        self._url = url
        self._email = email

    def build(self, request):
        contact_spec = {
            "name": self._name,
            "url": self._url,
            "email": self._email,
        }

        return {"contact": {k: v for k, v in contact_spec.items()}}


class OpenApiLicense(object):
    """License Object

    License information for the exposed API.

    :param name: REQUIRED. The license name used for the API.
    :type name: str
    :param url: A URL to the license used for the API. MUST be in the format
        of a URL.
    :type url: str
    """

    def __init__(self, name, url=None):
        super(OpenApiLicense, self).__init__()
        self._name = name
        self._url = url

    def build(self, request):
        license_spec = {
            "name": self._name,
        }

        if self._url:
            license_spec["url"] = self._url

        return {"license": license_spec}


class OpenApiApacheLicense(OpenApiLicense):
    def __init__(self):
        super(OpenApiApacheLicense, self).__init__(
            name="Apache 2.0",
            url="https://www.apache.org/licenses/LICENSE-2.0.html",
        )


class OpenApiInfo(object):
    """Info Object

    The object provides metadata about the API. The metadata MAY be used by
    the clients if needed, and MAY be presented in editing or documentation
    generation tools for convenience.


    :param title: REQUIRED. The title of the API.
    :type title: str
    :param version: REQUIRED. The version of the OpenAPI document (which is
        distinct from the OpenAPI Specification version or the API
        implementation version).
    :type version: str
    :param description: A short description of the API. CommonMark syntax
        (https://spec.commonmark.org/) MAY be used for rich text
        representation.
    :type description: str
    :param license: The license information for the exposed API.
    :type license: :class:`OpenApiLicense`
    :param terms_of_service: A URL to the Terms of Service for the API. MUST
        be in the format of a URL.
    :type terms_of_service: str
    :param contact: The contact information for the exposed API.

    """

    def __init__(
        self,
        title=None,
        version=None,
        description=None,
        license=None,
        terms_of_service=None,
        contact=None,
    ):
        super(OpenApiInfo, self).__init__()
        self._title = title or "OpenAPI service schema"
        self._version = version or "1.0.0"
        self._description = description or ""
        self._license = license
        self._terms_of_service = terms_of_service
        self._contact = contact

    def build(self, request):
        info_spec = {
            "title": self._title,
            "version": self._version,
        }

        if self._description:
            info_spec["description"] = self._description

        if self._terms_of_service:
            info_spec["termsOfService"] = self._terms_of_service

        if self._license:
            info_spec.update(self._license.build(request))
        if self._contact:
            info_spec.update(self._contact.build(request))

        return {"info": info_spec}


class OpenApiPaths(object):
    def __init__(self):
        super(OpenApiPaths, self).__init__()

    def _build_api_paths(self, route, current_path, request, parameters=None):
        result_spec = collections.defaultdict(dict)

        # process collection paths
        for http_method, ra_method in c.HTTP_TO_RA_COLLECTION_METHODS.items():
            if route.check_allow_methods(ra_method):
                if current_path == "/":
                    result_spec[current_path][http_method.lower()] = {
                        "summary": "Base application url",
                        "responses": oa_c.OPENAPI_FILTER_RESPONSE,
                        "operationId": "Get_urls",
                    }
                else:
                    result_spec[current_path][http_method.lower()] = {}
            # process next routes
            for route_name in route.get_routes():
                next_route = route.get_route(route_name)
                if next_route.is_collection_route():
                    next_path = utils.lastslash(
                        posixpath.join(current_path, route_name)
                    )
                    result_spec.update(
                        self._build_api_paths(
                            next_route, next_path, request, parameters
                        )
                    )
                    paths, schemas = next_route(request).build_openapi_specification(
                        next_path, parameters
                    )
                    result_spec.update(paths)

        return result_spec

    def build(self, request, parameters=None):
        paths_spec = {}

        main_route = request.application.main_route
        paths_spec.update(self._build_api_paths(main_route, "/", request, parameters))

        return {"paths": paths_spec}


class OpenApiComponents(object):
    def __init__(self):
        super(OpenApiComponents, self).__init__()

    @staticmethod
    def _build_schemas_by_resources(route, schema_generator):
        schemas = {"schemas": {"Error": oa_c.ERROR_SCHEMA}}
        for method in route.get_allow_methods():
            if method == c.DELETE:
                continue
            schema_name = schema_generator.resource_method_name(method)
            schemas["schemas"][schema_name] = schema_generator.generate_schema_object(
                method
            )
        return schemas

    @staticmethod
    def _build_parameters_by_resources(schema_generator, request):
        return {
            "parameters": schema_generator.generate_parameter_object(request),
        }

    @staticmethod
    def _build_responses():
        return {"responses": {"Error": oa_c.ERROR_RESPONSE}}

    @staticmethod
    def _merge_specs(in_spec, from_spec):
        for key, value in from_spec.items():
            in_spec[key].update(value)
        return in_spec

    def _build_api_resources(self, route, request):
        resources_spec = collections.defaultdict(dict)

        self._merge_specs(resources_spec, self._build_responses())

        resource = route.get_controller(request).get_resource()
        schema_generator = oa_utils.ResourceSchemaGenerator(resource, route)
        if resource:
            resources_spec = self._merge_specs(
                in_spec=resources_spec,
                from_spec=self._build_schemas_by_resources(
                    route=route,
                    schema_generator=schema_generator,
                ),
            )
            resources_spec = self._merge_specs(
                in_spec=resources_spec,
                from_spec=self._build_parameters_by_resources(
                    schema_generator=schema_generator, request=request
                ),
            )

        for route_name in route.get_routes():
            next_route = route.get_route(route_name)(request)
            resources_spec = self._merge_specs(
                in_spec=resources_spec,
                from_spec=self._build_api_resources(next_route, request),
            )

        return resources_spec

    def build(self, request):
        main_route = request.application.main_route
        component_spec = self._build_api_resources(main_route, request)

        return {"components": component_spec}


class OpenApiSecurity(object):
    def __init__(self):
        super(OpenApiSecurity, self).__init__()


class OpenApiTag(object):
    """Tag Object

    Adds metadata to a single tag that is used by the Operation Object. It is
    not mandatory to have a Tag Object per tag defined in the Operation Object
    instances.

    :param name: REQUIRED. The name of the tag.
    :param description: A short description for the tag. CommonMark syntax MAY
        be used for rich text representation.
    :param external_docs: Additional external documentation for this tag.
    """

    def __init__(self, name, description="", external_docs=None):
        super(OpenApiTag, self).__init__()
        self._name = name
        self._description = description
        self._external_docs = external_docs

    @property
    def name(self):
        return self._name

    def build(self, request=None):
        value = {"name": self._name, "description": self._description}
        if self._external_docs:
            if isinstance(self._external_docs, OpenApiExternalDocs):
                value["externalDocs"] = self._external_docs.build(request)
        return value


class OpenApiTags(object):
    """Tags Object"""

    def __init__(self, tags):
        super(OpenApiTags, self).__init__()
        self._tags = tags

    def _build_route_tag(self, route, current_path, request):
        result_spec = list()

        # process next routes
        for route_name in route.get_routes():
            next_route = route.get_route(route_name)
            if next_route.is_collection_route():
                next_path = posixpath.join(current_path, route_name)
                result_spec.extend(
                    self._build_route_tag(next_route, next_path, request)
                )
                tags = next_route(request).openapi_tags()
                result_spec.extend(tags)
            elif next_route.is_resource_route():
                next_path = posixpath.join(current_path, route_name)
                result_spec.extend(
                    self._build_route_tag(next_route, next_path, request)
                )
                tags = next_route(request).openapi_tags()
                result_spec.extend(tags)

        return result_spec

    def build(self, request):
        tags = {"tags": []}
        for tag in self._tags:
            if isinstance(tag, OpenApiTag):
                tags["tags"].append(tag.build(request))

        main_route = request.application.main_route
        tags["tags"].extend(self._build_route_tag(main_route, "/", request))
        tags["tags"] = [
            i for n, i in enumerate(tags["tags"]) if i not in tags["tags"][n + 1 :]
        ]
        return tags


class OpenApiExternalDocs(object):
    def __init__(self, url, description=None):
        super(OpenApiExternalDocs, self).__init__()
        self._url = url
        self._description = description

    def build(self, request=None):
        return {"url": self._url, "description": self._description}


class OpenApiServers(object):
    def __init__(self, url=None, description=None, variables=None, versions=None):
        super(OpenApiServers, self).__init__()
        self._url = url
        self._description = description
        self._variables = variables
        self._versions = versions

    def build_variables(self):
        enum = self._versions or [oa_c.API_VERSION_V1]
        return self._variables or {"version": {"enum": enum, "default": enum[0]}}

    def build(self, request):
        servers = [
            {
                "url": self._url or request.host_url,
                "description": self._description or "",
                "variables": self.build_variables(),
            }
        ]
        return {"servers": servers}



# Copyright (c) 2014 Eugene Frolov <efrolov@mirantis.com>
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
# implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import sys

from oslo_config import cfg

from restalchemy import version


def parse(args):
    cfg.CONF(
        args=args,
        project="restalchemy",
        version="RESTAlchemy %s" % version.version_info,
    )
    return cfg.CONF.config_file


class ConfigFileIsntDefined(Exception):
    pass


def parse_or_die(args=[]):
    if not parse(args):
        raise ConfigFileIsntDefined()


def parse_sys_args_or_die():
    return parse_or_die(sys.argv[1:])



def make_test_name(testcase_func, param_num, param):
    """
    Generic name function for parameterized tests
    Example:
        test_smth_1_arg1.1_arg2.1
        test_smth_2_arg1.2_arg2.1
        test_smth_3_arg1.3_arg2.1
                  ^^^^^^^^^^^^^^^
    """
    params_str = "_".join(str(i).replace(" ", "_") for i in param[0])
    return f"{testcase_func.__name__}_{param_num}_{params_str}"



# Copyright 2021 Eugene Frolov.
#
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import unittest

from restalchemy.dm import filters
from restalchemy.dm import models
from restalchemy.dm import properties
from restalchemy.dm import types
from restalchemy.storage.sql.dialect.query_builder import q
from restalchemy.tests import fixtures


class SimpleModel(models.ModelWithUUID):
    __tablename__ = "simple_table"

    field_str = properties.property(types.String())
    field_int = properties.property(types.Integer())
    field_bool = properties.property(types.Boolean())


# NOTE(efrolov): Sort model properties for correct ordering in asserts
SimpleModel.properties.sort_properties()


class MySQLQueryBuilderTestCase(unittest.TestCase):
    def setUp(self):
        super(MySQLQueryBuilderTestCase, self).setUp()
        self.Q = q.Q
        self.flt = filters.AND(
            {"field_bool": filters.EQ(True)},
            {"field_int": filters.EQ(0)},
            {"field_str": filters.EQ("FAKE_STR")},
        )

    def tearDown(self):
        super(MySQLQueryBuilderTestCase, self).tearDown()
        del self.Q

    def test_simple_select(self):
        result = self.Q.select(
            model=SimpleModel,
            session=fixtures.SessionFixture(),
        ).compile()

        self.assertEqual(
            "SELECT"
            " `t1`.`field_bool` AS `t1_field_bool`,"
            " `t1`.`field_int` AS `t1_field_int`,"
            " `t1`.`field_str` AS `t1_field_str`,"
            " `t1`.`uuid` AS `t1_uuid`"
            " FROM"
            " `simple_table` AS `t1`",
            result,
        )

    def test_select_with_1_filter(self):
        query = self.Q.select(
            model=SimpleModel,
            session=fixtures.SessionFixture(),
        ).where({"field_bool": filters.EQ(True)})

        result_expression = query.compile()
        self.assertEqual(
            "SELECT"
            " `t1`.`field_bool` AS `t1_field_bool`,"
            " `t1`.`field_int` AS `t1_field_int`,"
            " `t1`.`field_str` AS `t1_field_str`,"
            " `t1`.`uuid` AS `t1_uuid`"
            " FROM"
            " `simple_table` AS `t1` "
            "WHERE"
            " `t1`.`field_bool` = %s",
            result_expression,
        )

    def test_select_with_filters_with_and(self):
        query = self.Q.select(
            model=SimpleModel,
            session=fixtures.SessionFixture(),
        ).where(self.flt)

        result_expression = query.compile()
        result_values = query.values()

        self.assertEqual(
            "SELECT"
            " `t1`.`field_bool` AS `t1_field_bool`,"
            " `t1`.`field_int` AS `t1_field_int`,"
            " `t1`.`field_str` AS `t1_field_str`,"
            " `t1`.`uuid` AS `t1_uuid`"
            " FROM"
            " `simple_table` AS `t1` "
            "WHERE"
            " (`t1`.`field_bool` = %s AND"
            " `t1`.`field_int` = %s AND"
            " `t1`.`field_str` = %s)",
            result_expression,
        )
        self.assertEqual([True, 0, "FAKE_STR"], result_values)

    def test_select_with_filters_with_or(self):
        my_filter = filters.OR(
            filters.AND({"field_int": filters.LT(9)}),
            filters.AND({"field_int": filters.GE(1), "field_str": filters.IsNot(None)}),
        )
        query = self.Q.select(
            model=SimpleModel,
            session=fixtures.SessionFixture(),
        ).where(my_filter)

        result_expression = query.compile()
        self.assertEqual(
            "SELECT `t1`.`field_bool` AS `t1_field_bool`,"
            " `t1`.`field_int` AS `t1_field_int`,"
            " `t1`.`field_str` AS `t1_field_str`,"
            " `t1`.`uuid` AS `t1_uuid` "
            "FROM `simple_table` AS `t1` "
            "WHERE (`t1`.`field_int` < %s"
            " OR (`t1`.`field_int` >= %s AND `t1`.`field_str` IS NOT %s))",
            result_expression,
        )

    def test_select_with_empty_filter(self):
        query = self.Q.select(
            model=SimpleModel,
            session=fixtures.SessionFixture(),
        ).where({})

        result_expression = query.compile()

        self.assertEqual(
            "SELECT"
            " `t1`.`field_bool` AS `t1_field_bool`,"
            " `t1`.`field_int` AS `t1_field_int`,"
            " `t1`.`field_str` AS `t1_field_str`,"
            " `t1`.`uuid` AS `t1_uuid`"
            " FROM"
            " `simple_table` AS `t1`",
            result_expression,
        )

    def test_select_two_where_clause(self):
        second_filter = filters.AND({"field_str": filters.EQ("FAKE_STR_TWO")})
        query = (
            self.Q.select(
                model=SimpleModel,
                session=fixtures.SessionFixture(),
            )
            .where(self.flt)
            .where(second_filter)
        )

        result_expression = query.compile()
        result_values = query.values()

        self.assertEqual(
            "SELECT"
            " `t1`.`field_bool` AS `t1_field_bool`,"
            " `t1`.`field_int` AS `t1_field_int`,"
            " `t1`.`field_str` AS `t1_field_str`,"
            " `t1`.`uuid` AS `t1_uuid`"
            " FROM"
            " `simple_table` AS `t1` "
            "WHERE"
            " ((`t1`.`field_bool` = %s AND"
            " `t1`.`field_int` = %s AND"
            " `t1`.`field_str` = %s) AND"
            " `t1`.`field_str` = %s)",
            result_expression,
        )
        self.assertEqual([True, 0, "FAKE_STR", "FAKE_STR_TWO"], result_values)

    def test_select_with_filters_and_limit(self):
        query = (
            self.Q.select(
                model=SimpleModel,
                session=fixtures.SessionFixture(),
            )
            .where(self.flt)
            .limit(2)
        )

        result_expression = query.compile()
        result_values = query.values()

        self.assertEqual(
            "SELECT"
            " `t1`.`field_bool` AS `t1_field_bool`,"
            " `t1`.`field_int` AS `t1_field_int`,"
            " `t1`.`field_str` AS `t1_field_str`,"
            " `t1`.`uuid` AS `t1_uuid`"
            " FROM"
            " `simple_table` AS `t1` "
            "WHERE"
            " (`t1`.`field_bool` = %s AND"
            " `t1`.`field_int` = %s AND"
            " `t1`.`field_str` = %s) "
            "LIMIT 2",
            result_expression,
        )
        self.assertEqual([True, 0, "FAKE_STR"], result_values)

    def test_select_lock_with_filters(self):
        query = (
            self.Q.select(
                model=SimpleModel,
                session=fixtures.SessionFixture(),
            )
            .where(self.flt)
            .for_()
        )

        result_expression = query.compile()
        result_values = query.values()

        self.assertEqual(
            "SELECT"
            " `t1`.`field_bool` AS `t1_field_bool`,"
            " `t1`.`field_int` AS `t1_field_int`,"
            " `t1`.`field_str` AS `t1_field_str`,"
            " `t1`.`uuid` AS `t1_uuid`"
            " FROM"
            " `simple_table` AS `t1` "
            "WHERE"
            " (`t1`.`field_bool` = %s AND"
            " `t1`.`field_int` = %s AND"
            " `t1`.`field_str` = %s) "
            "FOR UPDATE",
            result_expression,
        )
        self.assertEqual([True, 0, "FAKE_STR"], result_values)

    def test_select_lock_with_filters_and_limit(self):
        query = (
            self.Q.select(
                model=SimpleModel,
                session=fixtures.SessionFixture(),
            )
            .where(self.flt)
            .for_()
            .limit(2)
        )

        result_expression = query.compile()
        result_values = query.values()

        self.assertEqual(
            "SELECT"
            " `t1`.`field_bool` AS `t1_field_bool`,"
            " `t1`.`field_int` AS `t1_field_int`,"
            " `t1`.`field_str` AS `t1_field_str`,"
            " `t1`.`uuid` AS `t1_uuid`"
            " FROM"
            " `simple_table` AS `t1` "
            "WHERE"
            " (`t1`.`field_bool` = %s AND"
            " `t1`.`field_int` = %s AND"
            " `t1`.`field_str` = %s) "
            "LIMIT 2 "
            "FOR UPDATE",
            result_expression,
        )
        self.assertEqual([True, 0, "FAKE_STR"], result_values)

    def test_select_order_by_with_filters(self):
        query = (
            self.Q.select(
                model=SimpleModel,
                session=fixtures.SessionFixture(),
            )
            .where(self.flt)
            .order_by("field_str")
        )
        query = query.order_by("field_int", "DESC")

        result_expression = query.compile()
        result_values = query.values()

        self.assertEqual(
            "SELECT"
            " `t1`.`field_bool` AS `t1_field_bool`,"
            " `t1`.`field_int` AS `t1_field_int`,"
            " `t1`.`field_str` AS `t1_field_str`,"
            " `t1`.`uuid` AS `t1_uuid`"
            " FROM"
            " `simple_table` AS `t1` "
            "WHERE"
            " (`t1`.`field_bool` = %s AND"
            " `t1`.`field_int` = %s AND"
            " `t1`.`field_str` = %s) "
            "ORDER BY"
            " `t1`.`field_str` ASC,"
            " `t1`.`field_int` DESC",
            result_expression,
        )
        self.assertEqual([True, 0, "FAKE_STR"], result_values)


class MySQLResultParserTestCase(unittest.TestCase):
    def test_simple_model_result_parser(self):
        row_from_db = {
            "t1_field_bool": "FakeBool",
            "t1_field_int": "FakeInt",
            "t1_field_str": "FakeStr",
            "t1_uuid": "FakeUUID",
        }
        select_clause = q.Q.select(
            SimpleModel,
            session=fixtures.SessionFixture(),
        )

        result = select_clause.parse_row(row_from_db)

        self.assertEqual(
            {
                "field_bool": "FakeBool",
                "field_int": "FakeInt",
                "field_str": "FakeStr",
                "uuid": "FakeUUID",
            },
            result,
        )



# Copyright 2019 Eugene Frolov
#
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.


from restalchemy.api import middlewares

from restalchemy.common import contexts


class ContextMiddleware(middlewares.Middleware):
    def __init__(
        self, application, context_class=contexts.Context, context_kwargs=None
    ):
        """
        Initialize the middleware with a context class.

        :param application: The next application down the WSGI stack.
        :type application: callable
        :param context_class: The class used to construct a request context.
        :type context_class: restalchemy.common.contexts.Context
        :param conext_kwargs: Additional keyword arguments to pass to the
            context class constructor.
        :type conext_kwargs: dict
        """
        super().__init__(application)
        self._context_class = context_class
        self._context_kwargs = context_kwargs or {}

    def _construct_context(self, req):
        """
        Constructs a context for the given request.

        This method initializes and returns an instance of the context
        class specified during the middleware initialization. The context
        is used to manage request-specific state and operations.

        :param req: The request object for which the context is being
            constructed.
        :return: An instance of the context class.
        """

        return self._context_class(**self._context_kwargs)

    def _get_response(self, ctx, req):
        """Call next application down the stack and return response.

        If this returns None, the next application down the stack will be
        executed. If it returns a response then that response will be returned
        and execution will stop here.

        :param ctx: The context object.
        :param req: The request object.
        :return: The response object.
        """
        return req.get_response(self.application)

    def process_request(self, req):
        """Called on each request.

        If this returns None, the next application down the stack will be
        executed. If it returns a response then that response will be returned
        and execution will stop here.

        :param req: The request object.
        :return: The response object.
        """
        ctx = self._construct_context(req)
        req.context = ctx
        with ctx.session_manager():
            return self._get_response(ctx, req)



#    Copyright 2025 Genesis Corporation.
#
#    All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.


RA_SERVICE_NAME = "ra_test"


RA_POSTGRESQL_PROTO_NAME = "postgresql"
RA_POSTGRESQL_DB_PORT = 5432
RA_POSTGRESQL_DB_HOST = "127.0.0.1"
RA_POSTGRESQL_USERNAME = "postgres"
RA_POSTGRESQL_PASSWORD = "password"
RA_POSTGRESQL_DB_NAME = RA_SERVICE_NAME


RA_MYSQL_PROTO_NAME = "mysql"
RA_MYSQL_DB_PORT = 3306
RA_MYSQL_DB_HOST = "127.0.0.1"
RA_MYSQL_USERNAME = "root"
RA_MYSQL_PASSWORD = "password"
RA_MYSQL_DB_NAME = RA_SERVICE_NAME

DB_CONFIG_SECTION = "db"
